from typing import Optional
from anndata import AnnData
import numpy as np
from scipy.sparse import csr_matrix
import pandas as pd
from ._weighting_matrix import (
    calculate_weight_matrix,
    impute_neighbour,
    _WEIGHTING_MATRIX,
    _PLATFORM,
)


def SME_normalize(
    adata: AnnData,
    use_data: str = "raw",
    weights: _WEIGHTING_MATRIX = "weights_matrix_all",
    platform: _PLATFORM = "Visium",
    copy: bool = False,
) -> Optional[AnnData]:
    """\
    using spatial location (S), tissue morphological feature (M) and gene expression (E) information to normalize data.

    Parameters
    ----------
    adata
        Annotated data matrix.
    use_data
        Input data, can be `raw` counts or log transformed data
    weights
        Weighting matrix for imputation.
        if `weights_matrix_all`, matrix combined all information from spatial location (S),
        tissue morphological feature (M) and gene expression (E)
        if `weights_matrix_pd_md`, matrix combined information from spatial location (S),
        tissue morphological feature (M)
    platform
        `Visium` or `Old_ST`
    copy
        Return a copy instead of writing to adata.
    Returns
    -------
    Anndata
    """
    if use_data == "raw":
        if isinstance(adata.X, csr_matrix):
            count_embed = adata.X.toarray()
        elif isinstance(adata.X, np.ndarray):
            count_embed = adata.X
        elif isinstance(adata.X, pd.Dataframe):
            count_embed = adata.X.values
        else:
            raise ValueError(
                f"""\
                    {type(adata.X)} is not a valid type.
                    """
            )
    else:
        count_embed = adata.obsm[use_data]

    calculate_weight_matrix(adata, platform=platform)

    impute_neighbour(adata, count_embed=count_embed, weights=weights)

    imputed_data = adata.obsm["imputed_data"].astype(float)
    imputed_data[imputed_data == 0] = np.nan
    adjusted_count_matrix = np.nanmean(np.array([count_embed, imputed_data]), axis=0)

    key_added = use_data + "_SME_normalized"
    adata.obsm[key_added] = adjusted_count_matrix

    print("The data adjusted by SME is added to adata.obsm['" + key_added + "']")

    return adata if copy else None



#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages

with open("README.md", encoding="utf8") as readme_file:
    readme = readme_file.read()

with open("HISTORY.rst") as history_file:
    history = history_file.read()

with open("requirements.txt") as f:
    requirements = f.read().splitlines()


setup_requirements = []

test_requirements = []

setup(
    author="Genomics and Machine Learning lab",
    author_email="duy.pham@uq.edu.au",
    python_requires=">=3.7",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    description="A downstream analysis toolkit for Spatial Transcriptomic data",
    entry_points={
        "console_scripts": [
            "stlearn=stlearn.app.cli:main",
        ],
    },
    install_requires=requirements,
    license="BSD license",
    long_description=readme + "\n\n" + history,
    long_description_content_type="text/markdown",
    include_package_data=True,
    keywords="stlearn",
    name="stlearn",
    packages=find_packages(include=["stlearn", "stlearn.*"]),
    setup_requires=setup_requirements,
    test_suite="tests",
    tests_require=test_requirements,
    url="https://github.com/BiomedicalMachineLearning/stLearn",
    version="0.4.9",
    zip_safe=False,
)



"""
Plotting of continuous features stored in adata.obs.
"""

from matplotlib import pyplot as plt
from PIL import Image
import pandas as pd
import matplotlib
import numpy as np

from typing import Optional, Union, Mapping  # Special
from typing import Sequence, Iterable  # ABCs
from typing import Tuple  # Classes

from anndata import AnnData
import warnings

from stlearn.plotting.classes import FeaturePlot
from stlearn.plotting.classes_bokeh import BokehGenePlot
from stlearn.plotting._docs import doc_spatial_base_plot, doc_gene_plot
from stlearn.utils import Empty, _empty, _AxesSubplot, _docs_params

from bokeh.io import push_notebook, output_notebook
from bokeh.plotting import show

# @_docs_params(spatial_base_plot=doc_spatial_base_plot, gene_plot=doc_gene_plot)
def feat_plot(
    adata: AnnData,
    feature: str = None,
    threshold: Optional[float] = None,
    contour: bool = False,
    step_size: Optional[int] = None,
    title: Optional["str"] = None,
    figsize: Optional[Tuple[float, float]] = None,
    cmap: Optional[str] = "Spectral_r",
    use_label: Optional[str] = None,
    list_clusters: Optional[list] = None,
    ax: Optional[matplotlib.axes._subplots.Axes] = None,
    fig: Optional[matplotlib.figure.Figure] = None,
    show_plot: Optional[bool] = True,
    show_axis: Optional[bool] = False,
    show_image: Optional[bool] = True,
    show_color_bar: Optional[bool] = True,
    color_bar_label: Optional[str] = "",
    zoom_coord: Optional[float] = None,
    crop: Optional[bool] = True,
    margin: Optional[bool] = 100,
    size: Optional[float] = 7,
    image_alpha: Optional[float] = 1.0,
    cell_alpha: Optional[float] = 0.7,
    use_raw: Optional[bool] = False,
    fname: Optional[str] = None,
    dpi: Optional[int] = 120,
    vmin: Optional[float] = None,
    vmax: Optional[float] = None,
) -> Optional[AnnData]:
    """\
    Allows the visualization of a continuous features stored in adata.obs
     for Spatial transcriptomics array.


    Parameters
    -------------------------------------
    {spatial_base_plot}
    {feature_plot}

    Examples
    -------------------------------------
    >>> import stlearn as st
    >>> adata = st.datasets.example_bcba()
    >>> st.pl.gene_plot(adata, 'dpt_pseudotime')

    """
    FeaturePlot(
        adata,
        feature=feature,
        threshold=threshold,
        contour=contour,
        step_size=step_size,
        title=title,
        figsize=figsize,
        cmap=cmap,
        use_label=use_label,
        list_clusters=list_clusters,
        ax=ax,
        fig=fig,
        show_plot=show_plot,
        show_axis=show_axis,
        show_image=show_image,
        show_color_bar=show_color_bar,
        color_bar_label=color_bar_label,
        zoom_coord=zoom_coord,
        crop=crop,
        margin=margin,
        size=size,
        image_alpha=image_alpha,
        cell_alpha=cell_alpha,
        use_raw=use_raw,
        fname=fname,
        dpi=dpi,
        vmin=vmin,
        vmax=vmax,
    )



"""
Functions for annotating spots as cell types.
"""

import os
import numpy as np
import pandas as pd
import scanpy as sc

import stlearn.tools.microenv.cci.r_helpers as rhs


def run_label_transfer(
    st_data, sc_data, sc_label_col, r_path, st_label_col=None, n_highly_variable=2000
):
    """Runs Seurat label transfer."""
    st_label_col = sc_label_col if type(st_label_col) == type(None) else st_label_col

    # Setting up the R environment #
    rhs.rpy2_setup(r_path)

    # Adding the source R code #
    r = rhs.ro.r
    path = os.path.dirname(os.path.realpath(__file__))
    r["source"](path + "/label_transfer.R")

    # Loading the label_transfer function #
    label_transfer_r = rhs.ro.globalenv["label_transfer"]
    print("Finished sourcing R code.")

    # Getting common gene set #
    sc_genes = sc_data.var_names
    st_genes = st_data.var_names
    genes = [gene for gene in sc_genes if gene in st_genes]
    sc_data = sc_data[:, genes].copy()
    st_data = st_data[:, genes].copy()

    # Extracting top & subsetting to top features to increase run-time speed #
    sc.pp.highly_variable_genes(
        st_data, n_top_genes=n_highly_variable, flavor="seurat_v3"
    )
    sc.pp.highly_variable_genes(
        sc_data, n_top_genes=n_highly_variable, flavor="seurat_v3"
    )
    genes_bool = np.logical_or(
        sc_data.var["highly_variable"].values, st_data.var["highly_variable"].values
    )
    sc_data = sc_data[:, genes_bool]
    st_data = st_data[:, genes_bool]
    print(
        f"Finished selecting & subsetting to hvgs: sc shape: "
        f"{sc_data.shape}, st shape: {st_data.shape}"
    )

    # Extracting the relevant information from anndatas #
    st_expr_df = st_data.to_df().transpose()
    sc_expr_df = sc_data.to_df().transpose()
    sc_labels = sc_data.obs[sc_label_col].values.astype(str)
    print(f"Finished extracting data. st shape: {st_expr_df.shape}")

    # R conversion of the data #
    sc_labels_r = rhs.ro.StrVector(sc_labels)
    with rhs.localconverter(rhs.ro.default_converter + rhs.pandas2ri.converter):
        # st_expr_df_r = rhs.ro.conversion.rpy2py(go_results_r)
        st_expr_df_r = rhs.ro.conversion.py2rpy(st_expr_df)
        sc_expr_df_r = rhs.ro.conversion.py2rpy(sc_expr_df)
    print("Finished py->rpy conversion.")

    # Running label transfer #
    label_transfer_scores_r = label_transfer_r(st_expr_df_r, sc_expr_df_r, sc_labels_r)
    print("Finished label transfer.")
    with rhs.localconverter(rhs.ro.default_converter + rhs.pandas2ri.converter):
        label_transfer_scores = rhs.ro.conversion.rpy2py(label_transfer_scores_r)
    print("Finished results rpy->py conversion.")

    # Adding to spatial anndata object #
    label_transfer_scores = label_transfer_scores.transpose()
    argmaxs = np.argmax(label_transfer_scores.values, axis=1)
    labels = [label_transfer_scores.columns.values[argmax] for argmax in argmaxs]

    st_data.obs[st_label_col] = labels
    st_data.obs[st_label_col] = st_data.obs[st_label_col].astype("category")
    label_transfer_scores.index = st_data.obs_names.values
    st_data.uns[st_label_col] = label_transfer_scores

    print(f"Spot labels added to st_data.obs[{st_label_col}].")
    print(f"Spot label scores added to st_data.uns[{st_label_col}].")


def get_counts(data):
    """Gets count data from anndata if available."""
    # Standard layer has counts #
    if type(data.X) != np.ndarray and np.all(np.mod(data.X[0, :].todense(), 1) == 0):
        counts = data.to_df().transpose()
    elif type(data.X) == np.ndarray and np.all(np.mod(data.X[0, :], 1) == 0):
        counts = data.to_df().transpose()
    elif (
        type(data.X) != np.ndarray
        and hasattr(data, "raw")
        and np.all(np.mod(data.raw.X[0, :].todense(), 1) == 0)
    ):
        counts = data.raw.to_adata()[data.obs_names, data.var_names].to_df().transpose()
    elif (
        type(data.X) == np.ndarray
        and hasattr(data, "raw")
        and np.all(np.mod(data.raw.X[0, :], 1) == 0)
    ):
        counts = data.raw.to_adata()[data.obs_names, data.var_names].to_df().transpose()
    else:
        raise Exception(
            "Inputted AnnData dosn't contain counts, only normalised "
            "values. Recreate with counts in .X or .raw.X."
        )

    return counts


def run_rctd(
    st_data,
    sc_data,
    sc_label_col,
    r_path,
    st_label_col=None,
    n_highly_variable=5000,
    min_cells=10,
    doublet_mode="full",
    n_cores=1,
):
    """Runs RCTD for deconvolution."""
    st_label_col = sc_label_col if type(st_label_col) == type(None) else st_label_col

    ########### Setting up the R environment #############
    rhs.rpy2_setup(r_path)

    # Adding the source R code #
    r = rhs.ro.r
    path = os.path.dirname(os.path.realpath(__file__))
    r["source"](path + "/rctd.R")

    # Loading the label_transfer function #
    rctd_r = rhs.ro.globalenv["rctd"]
    print("Finished sourcing R code.")

    # Getting common gene set #
    sc_genes = sc_data.var_names
    st_genes = st_data.var_names
    genes = [gene for gene in sc_genes if gene in st_genes]
    st_data_orig = st_data
    sc_data = sc_data[:, genes].copy()
    st_data = st_data[:, genes].copy()

    # Extracting top & subsetting to top features to increase run-time speed #
    sc.pp.highly_variable_genes(
        st_data, n_top_genes=n_highly_variable, flavor="seurat_v3"
    )
    sc.pp.highly_variable_genes(
        sc_data, n_top_genes=n_highly_variable, flavor="seurat_v3"
    )
    genes_bool = np.logical_or(
        sc_data.var["highly_variable"].values, st_data.var["highly_variable"].values
    )

    ###### Getting the count data (if available) ############
    st_counts = get_counts(st_data)
    sc_counts = get_counts(sc_data)

    st_counts = st_counts.loc[genes_bool, :]
    sc_counts = sc_counts.loc[genes_bool, :]

    st_coords = st_data.obs.loc[:, ["imagecol", "imagerow"]]
    sc_labels = sc_data.obs[sc_label_col].values.astype(str)
    print(f"Finished extracting counts data.")

    ####### Converting to R objects #########
    sc_labels_r = rhs.ro.StrVector(sc_labels)
    with rhs.localconverter(rhs.ro.default_converter + rhs.pandas2ri.converter):
        st_coords_r = rhs.ro.conversion.py2rpy(st_coords)
        st_counts_r = rhs.ro.conversion.py2rpy(st_counts)
        sc_counts_r = rhs.ro.conversion.py2rpy(sc_counts)
    print("Finished py->rpy conversion.")

    ######## Running RCTD ##########
    print("Running RCTD...")
    rctd_proportions_r = rctd_r(
        st_counts_r,
        st_coords_r,
        sc_counts_r,
        sc_labels_r,
        doublet_mode,
        min_cells,
        n_cores,
    )
    print("Finished RCTD deconvolution.")
    with rhs.localconverter(rhs.ro.default_converter + rhs.pandas2ri.converter):
        rctd_proportions = rhs.ro.conversion.rpy2py(rctd_proportions_r)
    print("Finished results rpy->py conversion.")

    # Adding to spatial anndata object #
    argmaxs = np.argmax(rctd_proportions.values, axis=1)
    labels = [rctd_proportions.columns.values[argmax] for argmax in argmaxs]
    st_data_orig.obs[st_label_col] = labels
    st_data_orig.obs[st_label_col] = st_data_orig.obs[st_label_col].astype("category")
    st_data_orig.uns[st_label_col] = rctd_proportions.loc[
        st_data_orig.obs_names.values, :
    ]

    print(f"Spot labels added to st_data.obs[{st_label_col}].")
    print(f"Spot label scores added to st_data.uns[{st_label_col}].")


def run_singleR(
    st_data,
    sc_data,
    sc_label_col,
    r_path,
    st_label_col=None,
    n_highly_variable=5000,
    n_centers=3,
    de_n=200,
    de_method="t",
):
    """Runs SingleR spot annotation."""
    st_label_col = sc_label_col if type(st_label_col) == type(None) else st_label_col
    ########### Setting up the R environment #############
    rhs.rpy2_setup(r_path)

    # Adding the source R code #
    r = rhs.ro.r
    path = os.path.dirname(os.path.realpath(__file__))
    r["source"](path + "/singleR.R")

    # Loading the label_transfer function #
    singleR_r = rhs.ro.globalenv["singleR"]
    print("Finished sourcing R code.")

    # Getting common gene set #
    sc_genes = sc_data.var_names
    st_genes = st_data.var_names
    genes = [gene for gene in sc_genes if gene in st_genes]
    st_data_orig = st_data
    sc_data = sc_data[:, genes].copy()
    st_data = st_data[:, genes].copy()

    # Extracting top & subsetting to top features to increase run-time speed #
    sc.pp.highly_variable_genes(
        st_data, n_top_genes=n_highly_variable, flavor="seurat_v3"
    )
    sc.pp.highly_variable_genes(
        sc_data, n_top_genes=n_highly_variable, flavor="seurat_v3"
    )
    genes_bool = np.logical_or(
        sc_data.var["highly_variable"].values, st_data.var["highly_variable"].values
    )
    sc_data = sc_data[:, genes_bool]
    st_data = st_data[:, genes_bool]
    print(f"Finished selecting & subsetting to hvgs.")

    # Extracting the relevant information from anndatas #
    st_expr_df = st_data.to_df().transpose()
    sc_expr_df = sc_data.to_df().transpose()
    sc_labels = sc_data.obs[sc_label_col].values.astype(str)
    print(f"Finished extracting data.")

    # R conversion of the data #
    sc_labels_r = rhs.ro.StrVector(sc_labels)
    de_method_r = rhs.ro.StrVector(de_method)
    with rhs.localconverter(rhs.ro.default_converter + rhs.pandas2ri.converter):
        st_expr_df_r = rhs.ro.conversion.py2rpy(st_expr_df)
        sc_expr_df_r = rhs.ro.conversion.py2rpy(sc_expr_df)
    print("Finished py->rpy conversion.")

    # Running label transfer #
    singleR_scores_r = singleR_r(
        st_expr_df_r, sc_expr_df_r, sc_labels_r, n_centers, de_n, de_method_r
    )
    print("Finished SingleR annotation.")
    with rhs.localconverter(rhs.ro.default_converter + rhs.pandas2ri.converter):
        singleR_scores = rhs.ro.conversion.rpy2py(singleR_scores_r)
    print("Finished results rpy->py conversion.")

    # Adding the results to the object #
    singleR_scores.index = [index.replace(".", "-") for index in singleR_scores.index]
    st_data_orig.obs[st_label_col] = singleR_scores.loc[:, "labels"].astype("category")
    singleR_scores_only = singleR_scores.drop(columns=["labels"])
    st_data_orig.uns[st_label_col] = singleR_scores_only

    print(f"Spot labels added to st_data.obs[{st_label_col}].")
    print(f"Spot label scores added to st_data.uns[{st_label_col}].")



"""Purpose of this script is to create general forms that are programmable with
	particular input. Will impliment forms for subsetting the data and
	visualisation options in a general way so can be used with any
	SingleCellAnalysis dataset.
"""

import sys
from flask_wtf import FlaskForm

# from flask_wtf.file import FileField
from wtforms import SelectMultipleField, SelectField
import wtforms


def createSuperForm(elements, element_fields, element_values, validators=None):
    """ Creates a general form; goal is to create a fully programmable form \
	that essentially governs all the options the user will select.

	Args:
		elements (list<str>): Element names to be rendered on the page, in \
							  order of how they will appear on the page.

		element_fields (list<str>): The names of the fields to be rendered. \
								Each field is in same order as 'elements'. \
								Currently supported are: \
								'Title', 'SelectMultipleField', 'SelectField', \
								'StringField', 'Text', 'List'.

		element_values (list<object>): The information which will be put into \
									the field. Changes depending on field: \

									'Title' and 'Text': 'object' is a string
									containing the title which will be added as \
									a heading when rendered on the page.

									'SelectMultipleField' and 'SelectField':
									'object' is list of options to select from.

									'StringField':
									The example values to display within the \
									fields text area. The 'placeholder' option.

									'List':
									A list of objects which will be attached \
									to the form.

		validators (list<FunctionHandles>): A list of functions which take the \
						form as input, used to construct the form validator. \
						Form validator constructed by calling these \
						sequentially with form 'self' as input.

	Args:
		form (list<WTForm>): A WTForm which has attached as variable all the \
		fields mentioned, so then when rendered as input to
		'SuperDataDisplay.html' shows the form.
	"""

    class SuperForm(FlaskForm):
        """A base form on which all of the fields will be added."""

    if type(validators) == type(None):
        validators = [None] * len(elements)

    # Add the information #
    SuperForm.elements = elements
    SuperForm.element_fields = element_fields

    multiSelectLeft = True  # Places multi-select field to left, alternatives
    # if many multi-selects in row
    for i, element in enumerate(elements):
        fieldName = element_fields[i]

        # Adding each element as the appropriate field to the form #
        if fieldName == "SelectMultipleField":
            setattr(
                SuperForm,
                element,
                SelectMultipleField(element, choices=element_values[i]),
            )
            # The point of this number is to give an order for the attributes,
            # so that odd numbers get rendered to right of page, even numbers
            # left.
            setattr(SuperForm, element + "_number", int(multiSelectLeft))
            # inverts, so if left, goes right for the next multiSelectField
            multiSelectLeft = multiSelectLeft == False

        else:
            multiSelectLeft = True  # Reset the MultiSelectField position

            if fieldName in ["Title", "List"]:
                setattr(SuperForm, element, element_values[i])

            elif fieldName == "SelectField":
                setattr(
                    SuperForm,
                    element,
                    SelectField(
                        element, choices=element_values[i], validators=validators[i]
                    ),
                )

            # elif fieldName == 'FileField':
            # 	setattr(SuperForm, element, FileField(validators=validators[i]))
            # 	setattr(SuperForm, element + '_placeholder',  # Setting default
            # 			element_values[i])

            elif fieldName in [
                "StringField",
                "IntegerField",
                "BooleanField",
                "FileField",
                "FloatField",
            ]:
                FieldClass = getattr(wtforms, fieldName)
                setattr(
                    SuperForm, element, FieldClass(element, validators=validators[i])
                )
                setattr(
                    SuperForm,
                    element + "_placeholder",  # Setting default
                    element_values[i],
                )

    return SuperForm


def getPreprocessForm():
    """Gets the preprocessing form generated from the superform above.

    Returns:
            FlaskForm: With attributes that allow for inputs that are related to
                                    pre-processing.
    """
    elements = [
        "Spot Quality Control Filtering",  # Title
        "Minimum genes per spot",
        "Minimum counts per spot",
        "Gene Quality Control Filtering",  # Title
        "Minimum spots per gene",
        "Minimum counts per gene",
        "Normalisation, Log-transform, & Scaling",  # Title
        "Normalize total",
        "Log 1P",
        "Scale",
    ]
    element_fields = [
        "Title",
        "IntegerField",
        "IntegerField",
        "Title",
        "IntegerField",
        "IntegerField",
        "Title",
        "BooleanField",
        "BooleanField",
        "BooleanField",
    ]
    element_values = ["", 200, 300, "", 3, 5, "", True, True, True]
    return createSuperForm(elements, element_fields, element_values)


def getLRForm():
    """Gets the LR form generated from the superform above.

        Returns:
                FlaskForm: With attributes that allow for inputs that are \
                            related to LR analysis.
    """
    elements = [
        "Species",
        "Spot neighbourhood (-1: smallest neighbourhood, 0: within-spot mode)",
        "Minimum spots with LR scores",
        "N random gene pairs (permutations)",
        "CPUs",
    ]
    element_fields = [
        "SelectField",
        "IntegerField",
        "IntegerField",
        "IntegerField",
        "IntegerField",
    ]
    element_values = [
        [("Human", "Human"), ("Mouse", "Mouse")],
        -1,
        20,
        100,
        2,
    ]
    return createSuperForm(elements, element_fields, element_values)


def getCCIForm(adata):
    """Gets the CCI form generated from the superform above.

    Returns:
            FlaskForm: With attributes that allow for inputs that are
                                                    related to CCI analysis.
    """
    elements = [
        "Cell information (only discrete labels available, unless mixture already in anndata.uns)",
        "Minimum spots for LR to be considered",
        "Spot mixture (only if the 'Cell Information' label selected available in anndata.uns)",
        "Cell proportion cutoff (value above which cell is considered in spot if 'Spot mixture' selected)",
        "Permutations (recommend atleast 1000)",
    ]
    element_fields = [
        "SelectField",
        "IntegerField",
        "BooleanField",
        "FloatField",
        "IntegerField",
    ]
    if type(adata) == type(None):
        fields = []
        mix = False
    else:
        fields = [
            key for key in adata.obs.keys() if type(adata.obs[key].values[0]) == str
        ]
        mix = fields[0] in adata.uns.keys()
    element_values = [fields, 20, mix, 0.2, 100]
    return createSuperForm(elements, element_fields, element_values)


def getCCIForm_old():
    """Gets the CCI form generated from the superform above.

    Returns:
            FlaskForm: With attributes that allow for inputs that are related to
                                    CCI analysis.
    """
    elements = [
        "* Cell Heterogeneity File",
        "Neighbourhood distance (0 indicates within-spot mode)",
        "** L-R pair input (e.g. L1_R1, L2_R2, ...)",
        "Permutations (0 indicates no permutation testing)",
    ]
    element_fields = ["FileField", "IntegerField", "StringField", "IntegerField"]
    element_values = ["", 25, "", 0]
    return createSuperForm(elements, element_fields, element_values)


def getClusterForm():
    """Gets the Cluster form generated using superform above.

    Returns:
            FlaskForm: With attributes that allow input related to clustering.
    """
    elements = [
        "PCA components",
        "stSME normalisation",
        "Cluster method",
        "K",
        "Resolution",
        "Neighbours (for Louvain/Leiden)",
    ]
    element_fields = [
        "IntegerField",
        "BooleanField",
        "SelectField",
        "IntegerField",
        "FloatField",
        "IntegerField",
    ]
    element_values = [
        50,
        True,
        [("KMeans", "KMeans"), ("Louvain", "Louvain"), ("Leiden", "Leiden")],
        10,
        1.0,
        15,
    ]
    return createSuperForm(elements, element_fields, element_values)


def getPSTSForm(trajectory, clusts, options):
    """Gets the psts form generated using superform above.

    Args:
            cluster_set (numpy.array<str>): The clusters which can be selected as
                                                                            the root for psts analysis.

    Returns:
            FlaskForm: With attributes that allow input related to psts.
    """
    elements = [
        "Root cluster",
        "Reverse",
        "eps (max. dist. spot neighbourhood)",
        "Trajectory Select",
        "Select distance-based method",
    ]
    element_fields = [
        "SelectField",
        "BooleanField",
        "IntegerField",
        "SelectField",
        "SelectField",
    ]

    element_values = [clusts, False, 50, trajectory, options]
    return createSuperForm(elements, element_fields, element_values)


def getDEAForm(list_labels, methods):
    """Gets the psts form generated using superform above.

    Args:
            cluster_set (numpy.array<str>): The clusters which can be selected as
                                                                            the root for psts analysis.

    Returns:
            FlaskForm: With attributes that allow input related to psts.
    """
    elements = ["Use label", "Use method"]
    element_fields = [
        "SelectField",
        "SelectField",
    ]

    element_values = [list_labels, methods]
    return createSuperForm(elements, element_fields, element_values)


######################## Junk Code #############################################
# def getCCIForm(step_log):
# 	""" Gets the CCI form generated from the superform above.
#
# 	Returns:
# 		FlaskForm: With attributes that allow for inputs that are related to
# 					CCI analysis.
# 	"""
# 	elements, element_fields, element_values = [], [], []
# 	if type(step_log['cci_het']) == type(None):
# 		# Analysis type form version #
# 		analysis_elements = ['Cell Heterogeneity Information', # Title
# 							 'cci_het',
# 							 'Permutation Testing', # Title
# 							 'cci_perm']
# 		analysis_fields = ['Title', 'SelectField', 'Title', 'SelectField']
# 		label_transfer_options = ['Upload Cell Label Transfer',
# 								  'No Cell Label Transfer']
# 		permutation_options = ['With permutation testing',
# 							   'Without permutation testing']
# 		analysis_values = ['', label_transfer_options, '', permutation_options]
# 		elements += analysis_elements
# 		element_fields += analysis_fields
# 		element_values += analysis_values
#
# 	else:
# 		# Core elements regardless of CCI mode #
# 		elements += ['Neighbourhood distance',
# 					 'L-R pair input (e.g. L1_R1, L2_R2, ...)']
# 		element_fields += ['IntegerField', 'StringField']
# 		element_values += [5, '']
#
# 		if step_log['cci_perm']:
# 			# Including cell heterogeneity information #
# 			elements += ['Permutations']
# 			element_fields += ['IntegerField']
# 			element_values += [200]
#
# 	return createSuperForm(elements, element_fields, element_values, None)



import os
import scanpy as sc
from PIL import Image
import numpy as np


def read_test_data():
    """Reads in test data to run unit tests."""
    # Determining path of this file #
    path = os.path.dirname(os.path.realpath(__file__))
    adata = sc.read_h5ad(f"{path}/test_data/test_data.h5")
    im = Image.open(f"{path}/test_data/test_image.jpg")
    adata.uns["spatial"]["V1_Breast_Cancer_Block_A_Section_1"]["images"][
        "hires"
    ] = np.array(im)
    return adata



import numpy as np
import pandas as pd
from scipy.spatial.distance import euclidean, canberra
from sklearn.preprocessing import MinMaxScaler

from numba import njit, prange
from numba.typed import List

from .base import get_lrs_scores


def nonzero_quantile(expr, q, interpolation):
    """Calculating the non-zero quantiles."""
    nonzero_expr = expr[expr > 0]
    quants = np.quantile(nonzero_expr, q=q, interpolation=interpolation)
    if type(quants) != np.array and type(quants) != np.ndarray:
        quants = np.array([quants])
    return quants


def getzero_prop(expr):
    """Calculating the proportion of zeros."""
    zero_bool = expr == 0
    n_zeros = len(np.where(zero_bool)[0])
    zero_prop = [n_zeros / len(expr)]
    return zero_prop


def get_lr_quants(
    lr_expr: pd.DataFrame,
    l_indices: list,
    r_indices: list,
    quantiles: np.array,
    method="",
):
    """Gets the quantiles per gene in the LR pair, & then concatenates.
    Returns
    -------
    lr_quants, l_quants, r_quants: np.ndarray   First is concatenation of two latter. Each row is a quantile value, each column is a LR pair.
    """

    quant_func = nonzero_quantile if method != "quantiles" else np.quantile

    # First getting the quantiles of gene expression #
    gene_quants = np.apply_along_axis(
        quant_func, 0, lr_expr.values, q=quantiles, interpolation="nearest"
    )

    l_quants = gene_quants[:, l_indices]
    r_quants = gene_quants[:, r_indices]

    lr_quants = np.concatenate((l_quants, r_quants), 0).transpose()

    return lr_quants, l_quants, r_quants


def get_lr_zeroprops(lr_expr: pd.DataFrame, l_indices: list, r_indices: list):
    """Gets the proportion of zeros per gene in the LR pair, & then concatenates.
    Returns
    -------
    lr_props, l_props, r_props: np.ndarray   First is concatenation of two latter. Each row is a prop value, each column is a LR pair.
    """

    # First getting the quantiles of gene expression #
    gene_props = np.apply_along_axis(getzero_prop, 0, lr_expr.values)

    l_props = gene_props[:, l_indices]
    r_props = gene_props[:, r_indices]

    lr_props = np.concatenate((l_props, r_props), 0).transpose()

    return lr_props, l_props, r_props


def get_lr_bounds(lr_value: float, bin_bounds: np.array):
    """For the given lr_value, returns the bin where it belongs.
    Returns
    -------
    lr_bin: tuple   Tuple of length 2, first is the lower bound of the bin, second is upper bound of the bin.
    """
    if np.any(bin_bounds == lr_value):  # If sits on a boundary
        lr_i = np.where(bin_bounds == lr_value)[0][0]
        if lr_value == max(bin_bounds):  # Must be in the final bin
            _lower = bin_bounds[-2]
            _upper = bin_bounds[-1]
        else:  # In the lower bin
            _lower = bin_bounds[lr_i]
            _upper = bin_bounds[lr_i + 1]
    else:  # Bin where it's value sit in-between #
        _lower = bin_bounds[np.where(bin_bounds < lr_value)[0][-1]]
        _upper = bin_bounds[np.where(bin_bounds > lr_value)[0][0]]

    return (_lower, _upper)


def get_similar_genes(
    ref_quants: np.array,
    ref_props: np.array,
    n_genes: int,
    candidate_expr: np.ndarray,
    candidate_genes: np.array,
    quantiles=(0.5),  # (.5, .75, .85, .9, .95, .97, .98, .99, 1)
):
    """Gets genes with a similar expression distribution as the inputted gene,
        by measuring distance between the gene expression quantiles.
    Parameters
    ----------
    ref_quants: np.array     The pre-calculated quantiles.
    ref_props: np.array      The query zero proportions.
    n_genes: int            Number of equivalent genes to select.
    candidate_expr: np.ndarray  Expression of gene candidates (cells*genes).
    candidate_genes: np.array   Same as candidate_expr.shape[1], indicating gene names.
    quantiles: tuple    The quantile to use
    Returns
    -------
    similar_genes: np.array Array of strings for gene names.
    """
    if type(quantiles) == float:
        quantiles = np.array([quantiles])
    else:
        quantiles = np.array(quantiles)

    # Query quants #
    query_quants = np.apply_along_axis(
        nonzero_quantile, 0, candidate_expr, q=quantiles, interpolation="nearest"
    )

    # Need to min-max normalise so can take distance with the proportion #
    all_quants = np.concatenate((np.array([ref_quants]), query_quants), axis=1)
    scaler = MinMaxScaler()
    scaled_quants = scaler.fit_transform(all_quants.transpose()).transpose()
    ref_scaled = scaled_quants[:, 0]
    query_scaled = scaled_quants[:, 1:]

    # Query props #
    query_props = np.apply_along_axis(getzero_prop, 0, candidate_expr)

    # Concatenating to create the ref & query vals to match #
    ref_vals = np.array([ref_scaled[0], ref_props[0]])  # both between 0 & 1
    query_vals = np.concatenate((query_scaled, query_props))

    # Measuring distances from the desired gene #
    dists = np.apply_along_axis(canberra, 0, query_vals, ref_vals)

    # Retrieving desired number of genes #
    order = np.argsort(dists)
    similar_genes = candidate_genes[order[0:n_genes]]

    """ During debugging, plotting distribution of distances & selected genes.
    import matplotlib.pyplot as plt
    cutoff = dists[order[n_genes]]
    plt.hist(dists, bins=100)
    plt.vlines(cutoff, 0, 500, color='r')
    plt.show()
    """

    return similar_genes


def get_similar_genes_Quantiles(
    gene_expr: np.array,
    n_genes: int,
    candidate_quants: np.ndarray,
    candidate_genes: np.array,
    quantiles=(0.5, 0.75, 0.85, 0.9, 0.95, 0.97, 0.98, 0.99, 1),
):
    """Gets genes with a similar expression distribution as the inputted gene,
        by measuring distance between the gene expression quantiles.
    Parameters
    ----------
    gene_expr: np.array     Expression of the gene of interest, or, if the same length as quantiles, then assumes is the pre-calculated quantiles.
    n_genes: int            Number of equivalent genes to select.
    candidate_quants: np.ndarray  Expression quantiles of gene candidates (quantiles*genes).
    candidate_genes: np.array   Same as candidate_expr.shape[1], indicating gene names.
    quantiles: tuple    The quantile to use
    Returns
    -------
    similar_genes: np.array Array of strings for gene names.
    """

    if type(quantiles) == float:
        quantiles = np.array([quantiles])
    else:
        quantiles = np.array(quantiles)

    # Getting the quantiles for the gene #
    if len(gene_expr) != len(quantiles):
        # ref_quants = np.quantile(gene_expr, q=quantiles, interpolation='nearest')
        ref_quants = nonzero_quantile(gene_expr, q=quantiles, interpolation="nearest")
    else:
        ref_quants = gene_expr

    # Measuring distances from the desired gene #
    dists = np.apply_along_axis(canberra, 0, candidate_quants, ref_quants)
    order = np.argsort(dists)

    """ During debugging, plotting distribution of distances & selected genes.
    import matplotlib.pyplot as plt
    cutoff = dists[order[n_genes]]
    fig, ax = plt.subplots()
    ax.hist(dists[order[0:28000]], bins=1000)
    y_max = ax.get_ylim()[1]
    ax.vlines(cutoff, 0, y_max/2, color='r')
    plt.show()
    print(candidate_quants[:,order[0:3]]) # Showing the quantiles of selected
    print(candidate_quants[:,order[n_genes-3:n_genes]])
    print(ref_quants)
    """

    # Retrieving desired number of genes #
    similar_genes = candidate_genes[order[0:n_genes]]

    return similar_genes


@njit(parallel=True)
def get_similar_genesFAST(
    ref_quants: np.array,
    n_genes: int,
    candidate_quants: np.ndarray,
    candidate_genes: np.array,
):
    """Fast version of the above with parallelisation."""

    # Measuring distances from the desired gene #
    dists = np.zeros((1, candidate_quants.shape[1]), dtype=np.float64)[0, :]
    for i in prange(0, candidate_quants.shape[1]):
        cand_quants = candidate_quants[:, i]
        abs_diff = ref_quants - cand_quants
        abs_diff[abs_diff < 0] = -abs_diff[abs_diff < 0]
        dists[i] = np.nansum(abs_diff / (ref_quants + cand_quants))

    # Need to remove the zero-dists since this indicates they are expressed
    # exactly the same, & hence likely in the same spot !!!
    nonzero_bool = dists > 0
    dists = dists[nonzero_bool]
    candidate_quants = candidate_quants[:, nonzero_bool]
    candidate_genes = candidate_genes[nonzero_bool]
    order = np.argsort(dists)

    """ During debugging, plotting distribution of distances & selected genes.
    import matplotlib.pyplot as plt
    cutoff = dists[order[n_genes]]
    fig, ax = plt.subplots()
    ax.hist(dists[order[0:28000]], bins=1000)
    y_max = ax.get_ylim()[1]
    ax.vlines(cutoff, 0, y_max/2, color='r')
    plt.show()
    print(candidate_quants[:,order[0:3]]) # Showing the quantiles of selected
    print(candidate_quants[:,order[n_genes-3:n_genes]])
    print(ref_quants)
    """

    # Retrieving desired number of genes #
    similar_genes = candidate_genes[order[0:n_genes]]

    return similar_genes


@njit
def gen_rand_pairs(genes1: np.array, genes2: np.array, n_pairs: int):
    """Generates random pairs of genes."""

    rand_pairs = List()
    for j in range(0, n_pairs):
        l_rand = np.random.choice(genes1, 1)[0]
        r_rand = np.random.choice(genes2, 1)[0]
        rand_pair = "_".join([l_rand, r_rand])
        while rand_pair in rand_pairs or l_rand == r_rand:
            l_rand = np.random.choice(genes1, 1)[0]
            r_rand = np.random.choice(genes2, 1)[0]
            rand_pair = "_".join([l_rand, r_rand])

        rand_pairs.append(rand_pair)

    return rand_pairs


def get_lr_features(adata, lr_expr, lrs, quantiles):
    """Gets expression features of LR pairs; nonzero-median, zero-prop, quantiles."""
    quantiles = np.array(quantiles)

    # Determining indices of LR pairs #
    l_indices, r_indices = [], []
    for lr in lrs:
        l_, r_ = lr.split("_")
        l_indices.extend(np.where(lr_expr.columns.values == l_)[0])
        r_indices.extend(np.where(lr_expr.columns.values == r_)[0])

    # The nonzero median when quantiles=.5 #
    lr_quants, l_quants, r_quants = get_lr_quants(
        lr_expr, l_indices, r_indices, quantiles, method="quantiles"
    )

    # Calculating the zero proportions, for grouping based on median/zeros #
    lr_props, l_props, r_props = get_lr_zeroprops(lr_expr, l_indices, r_indices)

    ######## Getting lr features for later diagnostics #######
    lr_meds, l_meds, r_meds = get_lr_quants(
        lr_expr, l_indices, r_indices, quantiles=np.array([0.5]), method=""
    )
    lr_median_means = lr_meds.mean(axis=1)
    lr_prop_means = lr_props.mean(axis=1)

    # Calculating mean rank #
    median_order = np.argsort(lr_median_means)
    prop_order = np.argsort(lr_prop_means * -1)
    median_ranks = [np.where(median_order == i)[0][0] for i in range(len(lrs))]
    prop_ranks = [np.where(prop_order == i)[0][0] for i in range(len(lrs))]
    mean_ranks = np.array([median_ranks, prop_ranks]).mean(axis=0)

    # Saving the lrfeatures...
    cols = ["nonzero-median", "zero-prop", "median_rank", "prop_rank", "mean_rank"]
    lr_features = pd.DataFrame(index=lrs, columns=cols)
    lr_features.iloc[:, 0] = lr_median_means
    lr_features.iloc[:, 1] = lr_prop_means
    lr_features.iloc[:, 2] = np.array(median_ranks)
    lr_features.iloc[:, 3] = np.array(prop_ranks)
    lr_features.iloc[:, 4] = np.array(mean_ranks)
    lr_features = lr_features.iloc[np.argsort(mean_ranks), :]
    lr_cols = [f"L_{quant}" for quant in quantiles] + [
        f"R_{quant}" for quant in quantiles
    ]
    quant_df = pd.DataFrame(lr_quants, columns=lr_cols, index=lrs)
    lr_features = pd.concat((lr_features, quant_df), axis=1)
    adata.uns["lrfeatures"] = lr_features

    return lr_features


def get_lr_bg(
    adata,
    neighbours,
    het_vals,
    min_expr,
    lr_,
    lr_score,
    l_quant,
    r_quant,
    genes,
    candidate_quants,
    gene_bg_genes,
    n_genes,
    n_pairs,
):
    """Gets the LR-specific background & bg spot indices."""
    l_, r_ = lr_.split("_")
    if l_ not in gene_bg_genes:
        l_genes = get_similar_genesFAST(
            l_quant, n_genes, candidate_quants, genes  # group_l_props,
        )
        gene_bg_genes[l_] = l_genes
    else:
        l_genes = gene_bg_genes[l_]

    if r_ not in gene_bg_genes:
        r_genes = get_similar_genesFAST(
            r_quant, n_genes, candidate_quants, genes  # group_r_props,
        )
        gene_bg_genes[r_] = r_genes
    else:
        r_genes = gene_bg_genes[r_]

    rand_pairs = gen_rand_pairs(l_genes, r_genes, n_pairs)
    spot_indices = np.where(lr_score > 0)[0]

    background = get_lrs_scores(
        adata,
        rand_pairs,
        neighbours,
        het_vals,
        min_expr,
        filter_pairs=False,
        spot_indices=spot_indices,
    )

    return background, spot_indices



import click
from .. import __version__

import os


@click.group(
    name="stlearn",
    subcommand_metavar="COMMAND <args>",
    options_metavar="<options>",
    context_settings=dict(max_content_width=85, help_option_names=["-h", "--help"]),
)
@click.help_option("--help", "-h", help="Show this message and exit.")
@click.version_option(
    version=__version__,
    prog_name="stlearn",
    message="[%(prog)s] Version %(version)s",
    help="Show the software version and exit.",
)
def main():
    os._exit
    click.echo("Please run `stlearn launch` to start the web app")


@main.command(short_help="Launch the stlearn interactive app")
def launch():
    from .app import app

    try:
        app.run(host="0.0.0.0", port=5000, debug=True, use_reloader=False)
    except OSError as e:
        if e.errno == errno.EADDRINUSE:
            raise click.ClickException(
                "Port is in use, please specify an open port using the --port flag."
            ) from e
        raise



from typing import Optional, Union
from anndata import AnnData
from pathlib import Path
import os
import pandas as pd


def lr(
    adata: AnnData,
    db_filepath: str = None,
    sep: str = "\t",
    source: str = "connectomedb",
    copy: bool = False,
) -> Optional[AnnData]:
    """Add significant Ligand-Receptor pairs into AnnData object

    Parameters
    ----------
    adata: AnnData          The data object to add L-R info into
    db_filepath: str      The path to the CPDB results file
    sep: str                Separator of the CPDB results file
    source: str             Source of LR database (default: connectomedb, can also support 'cellphonedb')
    copy: bool              Copy flag indicating copy or direct edit

    Returns
    -------
    adata: AnnData          The data object that L-R added into

    """

    if source == "cellphonedb":
        cpdb = pd.read_csv(db_filepath, sep=sep)
        adata.uns["cpdb"] = cpdb
        lr = cpdb["interacting_pair"].to_list()
        lr2 = [i for i in lr if "complex" not in i]
        lr3 = [i for i in lr2 if " " not in i]
        lr4 = [i for i in lr3 if i.count("_") == 1]
        adata.uns["lr"] = lr4
        print("cpdb results added to adata.uns['cpdb']")
        print("Added ligand receptor pairs to adata.uns['lr'].")

    elif source == "connectomedb":
        ctdb = pd.read_csv(db_filepath, sep=sep, quotechar='"', encoding="latin1")
        adata.uns["lr"] = (
            ctdb["Ligand gene symbol"] + "_" + ctdb["Receptor gene symbol"]
        ).values.tolist()
        print("connectomedb results added to adata.uns['ctdb']")
        print("Added ligand receptor pairs to adata.uns['lr'].")

    return adata if copy else None



from matplotlib import pyplot as plt
from PIL import Image
import pandas as pd
import matplotlib
import numpy as np

from typing import Optional, Union, Mapping  # Special
from typing import Sequence, Iterable  # ABCs
from typing import Tuple  # Classes

from anndata import AnnData
import warnings

from stlearn.plotting.classes import GenePlot
from stlearn.plotting.classes_bokeh import BokehGenePlot
from stlearn.plotting._docs import doc_spatial_base_plot, doc_gene_plot
from stlearn.utils import Empty, _empty, _AxesSubplot, _docs_params

from bokeh.io import push_notebook, output_notebook
from bokeh.plotting import show


@_docs_params(spatial_base_plot=doc_spatial_base_plot, gene_plot=doc_gene_plot)
def gene_plot(
    adata: AnnData,
    gene_symbols: Union[str, list] = None,
    threshold: Optional[float] = None,
    method: str = "CumSum",
    contour: bool = False,
    step_size: Optional[int] = None,
    title: Optional["str"] = None,
    figsize: Optional[Tuple[float, float]] = None,
    cmap: Optional[str] = "Spectral_r",
    use_label: Optional[str] = None,
    list_clusters: Optional[list] = None,
    ax: Optional[matplotlib.axes._subplots.Axes] = None,
    fig: Optional[matplotlib.figure.Figure] = None,
    show_plot: Optional[bool] = True,
    show_axis: Optional[bool] = False,
    show_image: Optional[bool] = True,
    show_color_bar: Optional[bool] = True,
    color_bar_label: Optional[str] = "",
    zoom_coord: Optional[float] = None,
    crop: Optional[bool] = True,
    margin: Optional[bool] = 100,
    size: Optional[float] = 7,
    image_alpha: Optional[float] = 1.0,
    cell_alpha: Optional[float] = 0.7,
    use_raw: Optional[bool] = False,
    fname: Optional[str] = None,
    dpi: Optional[int] = 120,
    vmin: Optional[float] = None,
    vmax: Optional[float] = None,
) -> Optional[AnnData]:
    """\
    Allows the visualization of a single gene or multiple genes as the values
    of dot points or contour in the Spatial transcriptomics array.


    Parameters
    -------------------------------------
    {spatial_base_plot}
    {gene_plot}

    Examples
    -------------------------------------
    >>> import stlearn as st
    >>> adata = st.datasets.example_bcba()
    >>> genes = ["BRCA1","BRCA2"]
    >>> st.pl.gene_plot(adata, gene_symbols = genes)

    """
    GenePlot(
        adata,
        gene_symbols=gene_symbols,
        threshold=threshold,
        method=method,
        contour=contour,
        step_size=step_size,
        title=title,
        figsize=figsize,
        cmap=cmap,
        use_label=use_label,
        list_clusters=list_clusters,
        ax=ax,
        fig=fig,
        show_plot=show_plot,
        show_axis=show_axis,
        show_image=show_image,
        show_color_bar=show_color_bar,
        color_bar_label=color_bar_label,
        zoom_coord=zoom_coord,
        crop=crop,
        margin=margin,
        size=size,
        image_alpha=image_alpha,
        cell_alpha=cell_alpha,
        use_raw=use_raw,
        fname=fname,
        dpi=dpi,
        vmin=vmin,
        vmax=vmax,
    )


def gene_plot_interactive(adata: AnnData):
    bokeh_object = BokehGenePlot(adata)
    output_notebook()
    show(bokeh_object.app, notebook_handle=True)



from matplotlib import pyplot as plt
from matplotlib.axes import Axes
from matplotlib.figure import Figure
import matplotlib
import pandas as pd
import numpy as np
import networkx as nx
import math
import matplotlib.patches as patches
from numba.typed import List
import seaborn as sns
import sys
from anndata import AnnData
from typing import Optional, Union

from typing import Optional, Union, Mapping  # Special
from typing import Sequence, Iterable  # ABCs
from typing import Tuple  # Classes

import warnings

from .classes import CciPlot, LrResultPlot
from .classes_bokeh import BokehSpatialCciPlot, BokehLRPlot
from ._docs import doc_spatial_base_plot, doc_het_plot, doc_lr_plot
from ..utils import Empty, _empty, _AxesSubplot, _docs_params
from .utils import get_cmap, check_cmap, get_colors
from .cluster_plot import cluster_plot
from .deconvolution_plot import deconvolution_plot
from .gene_plot import gene_plot
from stlearn.plotting.utils import get_colors
import stlearn.plotting.cci_plot_helpers as cci_hs
from .cci_plot_helpers import (
    get_int_df,
    add_arrows,
    create_flat_df,
    _box_map,
    chordDiagram,
)
from scipy.stats import gaussian_kde

import importlib

importlib.reload(cci_hs)

from bokeh.io import push_notebook, output_notebook
from bokeh.plotting import show

#### Functions for visualising the overall LR results and diagnostics.


def lr_diagnostics(
    adata,
    highlight_lrs: list = None,
    n_top: int = None,
    color0: str = "turquoise",
    color1: str = "plum",
    figsize: tuple = (10, 4),
    lr_text_fp: dict = None,
    show: bool = True,
):
    """Diagnostic plot looking at relationship between technical features of lrs and lr rank.
        Two plots generated: left is the average of the median for nonzero
        expressing spots for both the ligand and the receptor on the y-axis, &
        LR-rank by no. of significant spots on the x-axis. Right is the average
        of the proportion of zeros for the ligand and receptor gene on teh y-axis.

    Parameters
    ----------
    adata: AnnData
        The data object on which st.tl.cci.run has been applied.
    highlight_lrs: list
        List of LRs to highlight, will add text and change point color for these LR pairs.
    n_top: int
        The number of LRs to display. If None shows all.
    color0: str
        The color of the nonzero-median scatter plot.
    lr_text_fp: dict
        Font dict for the LR text if highlight_lrs not None.
    axis_text_fp: dict
        Font dict for the axis text labels.
    Returns
    -------
    Figure, Axes
        Figure and axes of the plot, if show=False.
    """
    if type(n_top) == type(None):
        n_top = adata.uns["lr_summary"].shape[0]
    fig, axes = plt.subplots(ncols=2, figsize=figsize)
    cci_hs.lr_scatter(
        adata,
        "nonzero-median",
        highlight_lrs=highlight_lrs,
        n_top=n_top,
        color=color0,
        ax=axes[0],
        lr_text_fp=lr_text_fp,
        show=False,
    )
    cci_hs.lr_scatter(
        adata,
        "zero-prop",
        highlight_lrs=highlight_lrs,
        n_top=n_top,
        color=color1,
        ax=axes[1],
        lr_text_fp=lr_text_fp,
        show=False,
    )
    if show:
        plt.show()
    else:
        return fig, axes


def lr_summary(
    adata,
    n_top: int = 50,
    highlight_lrs: list = None,
    y: str = "n_spots_sig",
    color: str = "gold",
    figsize: tuple = None,
    highlight_color: str = "red",
    max_text: int = 50,
    lr_text_fp: dict = None,
    ax: Axes = None,
    show: bool = True,
):
    """Plotting the top LRs ranked by number of significant spots.

    Parameters
    ----------
    adata: AnnData
        The data object on which st.tl.cci.run has been applied.
    n_top: int
        The no. of LRs to plot.
    highlight_lrs: list
        A list of LRs to highlight on the plot, will added text and change color
        of points for these LRs. Useful for highlighting LRs of interest.
    y: str
        The way to rank the LRs, default is by the no. of signifcant spots,
        but can be any column in adata.uns['lr_summary'].
    color: str
        The color of the points.
    figsize: tuple
        Size of the figure; (width, height).
    highlight_color: str
        Only relevant if highlight_lrs specified; controls colour of LRs to
        highlight.
    max_text: int
        If the no. of n_top is above this limit, stop showing text to indicate
        the LR names. Allows to see global shape without crowding with LR name
        text.
    lr_text_fp: dict
        Matplotlib font dictionary specifying text details, eg fontsize.
    ax: Axes
        Axes on which to draw the scatter plot; if not inputted constructs own.
    show: bool
        Whether to show the plot, if False will return the ax.
    Returns
    _______
    Axes
        If show=False, returns the ax for additional modification.
    """
    allowed = ["n_spots", "n_spots_sig", "n_spots_sig_pval"]
    if y not in allowed:
        raise Exception(f"Got {y} for y; must be one of {allowed}")

    return cci_hs.lr_scatter(
        adata,
        y,
        n_top=n_top,
        color=color,
        show_all=n_top <= max_text,
        figsize=figsize,
        highlight_lrs=highlight_lrs,
        ax=ax,
        lr_text_fp=lr_text_fp,
        highlight_color=highlight_color,
        show=show,
    )


def lr_n_spots(
    adata,
    n_top: int = 100,
    font_dict: dict = None,
    xtick_dict: dict = None,
    bar_width: float = 1,
    max_text: int = 50,
    non_sig_color: str = "dodgerblue",
    sig_color: str = "springgreen",
    figsize: tuple = (6, 4),
    show_title: bool = True,
    show: bool = True,
):
    """Bar plot showing for each LR no. of sig versus non-sig spots.

    Parameters
    ----------
    adata: AnnData
        Data on which st.tl.cci.run has been applied.
    n_top: int
        The no. of LRs to display, ranked by adata.uns['lr_summary']
    font_dict: dict
        dictionary specifying matplotlib font parameters e.g. weight.
    xtick_dict: dict
        dictionary specifying matplotlib font parameters for x labels e.g. weight.
    bar_width: int
        Width of each bar in the bar plot.
    max_text: int
        If n_top exceeds this number, stop showing the LR text, since can cause
        crowding.
    non_sig_color: str
        Specifies the color for bar plot proportion indicating no. of non-sig
        spots.
    sig_color: str
        Specifies color for bars indicating the sig. spots counts.
    figsize: tuple
        Specifies figure dimensions.
    show_title: bool
        Whether to show title on outputted plot.
    show: bool
        Whether to show the plot; if false returns the figure & axes for further
        modification.
    Returns
    -------
    Fig, Axes
        Figure & axes with the plot draw on; only if show=False. Else None.
    """
    if type(font_dict) == type(None):
        font_dict = {"weight": "bold", "size": 12}
    if type(xtick_dict) == type(None):
        xtick_dict = {"fontweight": "bold", "rotation": 90, "size": 6}

    lrs = adata.uns["lr_summary"].index.values[0:n_top]
    n_sig = adata.uns["lr_summary"].loc[:, "n_spots_sig"].values
    n_non_sig = adata.uns["lr_summary"].loc[:, "n_spots"].values - n_sig
    rank = list(range(len(n_sig)))
    fig, ax = plt.subplots(figsize=figsize)
    ax.bar(rank[0:n_top], n_non_sig[0:n_top], bar_width, color=non_sig_color)
    ax.bar(
        rank[0:n_top],
        n_sig[0:n_top],
        bar_width,
        bottom=n_non_sig[0:n_top],
        color=sig_color,
    )
    ax.set_ylabel("n_spots", font_dict)
    ax.set_xlabel("LRs Ranked (n_spots_sig)", font_dict)
    if show_title:
        ax.set_title("Signficant and non-signficant spots per LR", font_dict)
    ax.legend(labels=["non-sig", "sig"], loc="upper right")
    if n_top <= max_text:
        ax.set_xticks(rank[0:n_top])
        ax.set_xticklabels(lrs, fontdict=xtick_dict)
    ax.spines["top"].set_visible(False)
    ax.spines["right"].set_visible(False)

    if show:
        plt.show()
    else:
        return fig, ax


def lr_go(
    adata,
    n_top: int = 20,
    highlight_go: list = None,
    figsize=(6, 4),
    rot: float = 50,
    lr_text_fp: dict = None,
    highlight_color: str = "yellow",
    max_text: int = 50,
    show: bool = True,
):
    """Plots the results from the LR GO analysis.

    Parameters
    ----------
    adata: AnnData
        Data object on which st.tl.cci.lr & st.tl.cci.run_lr_go has been called.
    n_top: int
        Specifies the no. of GO terms to show.
    highlight_go: list
        Names of GO terms to highlight in different color & text.
    figsize: tuple
        Size of the figure.
    rot: int
        Rotation of the text labels for the GO terms.
    lr_text_fp: dict
        Dictionary specifying matplotlib text params that control rendering of
        label on the points.
    highlight_color: str
        The color to plot the GO terms specified in highlight_go.
    max_text: int
        If n_top exceeds this number, stop showing the GO text, since can cause
        crowding.
    show: bool
        Whether to show the plot.
    """
    # Making sure LR GO has been run #
    if "lr_go" not in adata.uns:
        raise Exception("Need to run st.tl.cci.run_lr_go() first!")

    go_results = adata.uns["lr_go"]
    gos = go_results.loc[:, "Description"].values.astype(str)
    y = -np.log10(go_results.loc[:, "p.adjust"].values)
    sizes = go_results.loc[:, "Count"].values
    cci_hs.rank_scatter(
        gos[0:n_top],
        y[0:n_top],
        point_sizes=sizes[0:n_top],
        highlight_items=highlight_go,
        lr_text_fp=lr_text_fp,
        highlight_color=highlight_color,
        figsize=figsize,
        y_label="-log10(padjs)",
        x_label="GO Rank",
        height=6,
        color="deepskyblue",
        rot=rot,
        width_ratio=0.4,
        show=show,
        point_size_name="n-genes",
        show_all=n_top <= max_text,
    )


def cci_check(
    adata: AnnData,
    use_label: str,
    figsize=(16, 10),
    cell_label_size=20,
    axis_text_size=18,
    tick_size=14,
    show=True,
):
    """Checks relationship between no. of significant CCI-LR interactions and cell type frequency.

    Parameters
    ----------
    adata: AnnData
        Data on which st.tl.cci.run & st.tl.cci.run_cci has been performed.
    use_label: str
        The cell type label information used when running st.tl.cci.run_cci
    figsize: tuple
        Size of outputted figure.
    cell_label_size: int
        Size of the cell labels put on top of the bar chart.
    axis_text_size: int
        Size of the axis text.
    tick_size: int
        Size of the ticks displayed at bottom of chart.
    show: bool
        Whether to show the plot or not; if false returns figure & axes.
    Returns
    -------
    Figure, Ax1, Ax2
        The figure, axes for the barchart, and twin axes for the lineplot.
    """
    labels = adata.obs[use_label].values.astype(str)
    label_set = np.array(list(adata.obs[use_label].cat.categories))
    colors = get_colors(adata, use_label)
    xs = np.array(list(range(len(label_set))))
    int_dfs = adata.uns[f"per_lr_cci_{use_label}"]

    # Counting!!! #
    cell_counts = []  # Cell type frequencies
    cell_sigs = []  # Cell type significant interactions
    for j, label in enumerate(label_set):
        counts = sum(labels == label)
        cell_counts.append(counts)

        int_count = 0
        for lr in int_dfs:
            int_df = int_dfs[lr]
            label_index = np.where(int_df.index.values == label)[0][0]
            int_bool = int_df.values > 0
            int_count += sum(int_bool[label_index, :])
            int_count += sum(int_bool[:, label_index])
            # prevent double counts
            int_count -= int_bool[label_index, label_index]

        cell_sigs.append(int_count)

    cell_counts = np.array(cell_counts)
    cell_sigs = np.array(cell_sigs)
    order = np.argsort(cell_counts)
    cell_counts = cell_counts[order]
    cell_sigs = cell_sigs[order]
    colors = np.array(colors)[order]
    label_set = label_set[order]

    # Plotting bar plot #
    fig, ax = plt.subplots(figsize=figsize)
    ax.bar(xs, cell_counts, color=colors)
    text_dist = max(cell_counts) * 0.015
    fontdict = {"fontweight": "bold", "fontsize": cell_label_size}
    for j in range(len(xs)):
        ax.text(
            xs[j],
            cell_counts[j] + text_dist,
            label_set[j],
            rotation=90,
            fontdict=fontdict,
        )
    axis_text_fp = {"fontweight": "bold", "fontsize": axis_text_size}
    ax.set_ylabel("Cell type frequency", color="black", **axis_text_fp)
    ax.spines["top"].set_visible(False)
    ax.tick_params(labelsize=tick_size)
    ax.set_xlabel("Cell type rank", **axis_text_fp)

    # Line-plot of the interaction counts #
    ax2 = ax.twinx()
    ax2.set_ylabel("CCI-LR interactions", color="blue", **axis_text_fp)
    ax2.plot(xs, cell_sigs, color="blue", linewidth=2)
    ax2.tick_params(axis="y", labelcolor="blue", labelsize=tick_size)
    ax2.spines["top"].set_visible(False)
    ax2.tick_params(labelsize=tick_size)
    fig.tight_layout()

    if show:
        plt.show()
    else:
        return fig, ax, ax2


# Functions for visualisation the LR results per spot.
def lr_result_plot(
    adata: AnnData,
    use_lr: Optional["str"] = None,
    use_result: Optional["str"] = "lr_sig_scores",
    # plotting param
    title: Optional["str"] = None,
    figsize: Optional[Tuple[float, float]] = None,
    cmap: Optional[str] = "Spectral_r",
    ax: Optional[matplotlib.axes._subplots.Axes] = None,
    fig: Optional[matplotlib.figure.Figure] = None,
    show_plot: Optional[bool] = True,
    show_axis: Optional[bool] = False,
    show_image: Optional[bool] = True,
    show_color_bar: Optional[bool] = True,
    zoom_coord: Optional[float] = None,
    crop: Optional[bool] = True,
    margin: Optional[float] = 100,
    size: Optional[float] = 7,
    image_alpha: Optional[float] = 1.0,
    cell_alpha: Optional[float] = 1.0,
    use_raw: Optional[bool] = False,
    fname: Optional[str] = None,
    dpi: Optional[int] = 120,
    contour: bool = False,
    step_size: Optional[int] = None,
    vmin: float = None,
    vmax: float = None,
):
    """Plots the per spot statistics for given LR.

    Parameters
    ----------
    adata: AnnData
        Data on which st.tl.cci.run has been performed
    use_lr: str
        LR to show results for.
    use_result: str
        LR matrix in data.obsm; 'lr_scores', 'lr_sig_scores', 'p_vals', 'p_adjs'
        '-log10(p_adjs)'.
    title: str
        Plot title.
    figsize: tuple
        Figure size.
    cmap: str
        Color of points.
    ax: Axes
        Axes on which to plot.
    fig: Figure
        Figure associated with axes.
    show_plot: bool
        Whether to show plot or not.
    show_axis: bool
        Whether to show axis or not.
    show_image: bool
        Whether to plot the image.
    show_color_bar: bool
        Whether to show the color bar.
    crop: bool
        Whether to crop the image down to match spatial coordinates of gene
        spot present.
    margin: float
        Margin around the points for the image.
    size: float
        Size of the points.
    image_alpha: float
        Transparency of image.
    cell_alpha: float
        Transparency of points.
    use_raw: bool
        Whether to use adata.raw or not.
    fname: str
        Name of file to save plot to.
    dpi: int
        Plot saving quality.
    contour: bool
        Whether to plot as contour.
    step_size: int
        Step size of contour==True
    vmin: float
        Minimum value of scale bar.
    vmax: float
        Maximum value of scale bar.
    """
    LrResultPlot(
        adata,
        use_lr,
        use_result,
        # plotting param
        title,
        figsize,
        cmap,
        None,
        ax,
        fig,
        show_plot,
        show_axis,
        show_image,
        show_color_bar,
        crop,
        zoom_coord,
        margin,
        size,
        image_alpha,
        cell_alpha,
        use_raw,
        fname,
        dpi,
        # cci_rank param
        contour,
        step_size,
        vmin,
        vmax,
    )


# @_docs_params(het_plot=doc_lr_plot)
def lr_plot(
    adata: AnnData,
    lr: str,
    min_expr: float = 0,
    sig_spots=True,
    use_label: str = None,
    outer_mode: str = "continuous",
    l_cmap=None,
    r_cmap=None,
    lr_cmap=None,
    inner_cmap=None,
    inner_size_prop: float = 0.25,
    middle_size_prop: float = 0.5,
    outer_size_prop: float = 1,
    pt_scale: int = 100,
    title="",
    show_image: bool = True,
    show_arrows: bool = False,
    fig: Figure = None,
    ax: Axes = None,
    arrow_head_width: float = 4,
    arrow_width: float = 0.001,
    arrow_cmap: str = None,
    arrow_vmax: float = None,
    sig_cci: bool = False,
    lr_colors: dict = None,
    figsize: tuple = (6.4, 4.8),
    use_mix: bool = None,
    # plotting params
    **kwargs,
) -> Optional[AnnData]:
    """Creates different kinds of spatial visualisations for the LR analysis results.
        To see combinations of parameters refer to stLearn CCI tutorial.

    Parameters
    ----------
    adata: Anndata
        Data on which st.tl.cci.run has been performed; extra options unlocked
        below when have performed st.tl.cci.run_cci as well.
    lr: str
        The LR to display results for.
    min_expr: float
        The minimum expr above which LR considered expressed when plotting
        binary LR expression.
    sig_spots: bool
        Whether to subset to significant spots or not.
    use_label: str
        The cell type labels to use if plotting cell types.
    outer_mode: str
        The mode for the larger points when displaying LR expression;
        can either be 'binary' or 'continuous' or None.
        'Binary' discretizes each spot as expressing L, R, both, or neither.
        'Continuous' shows color gradient for levels of LR expression by plotting
        two points for each spot, the 'inner' point is the receptor expression
        levels, and the 'outer' point is the ligand expression level. None
        plots no ligand/receptor expression.
    l_cmap: str
        Cmap for coloring the ligand expression, only if
        outer_mode=='continuous'.
    r_cmap: str
        Cmap for coloring the receptor expression, only if
        outer_mode=='continuous'.
    lr_cmap: str
        Cmap for coloring coexpression.
    inner_cmap: str
        Cmap for the inner point if outer_mode is 'binary'.
    inner_size_prop: float
        Proportion of the inner point size when plotting to points for one spot.
        Scale of 0 to 1.
    middle_size_prop: float
        Controls size of middle point if specifying parameters that plot
        3 points per spot to display multiple information. Scale 0 to 1.
    outer_size_prop: float
        Point size of the outer point.
    pt_scale: float
        Overall size of point.
    title: str
        Title of figure.
    show_image: bool
        Whether to show the background image.
    show_arrows: bool
        Whether to plot arrows indicating interactions between spots.
    fig: Figure
        Figure to draw on.
    ax: Axes
        Axes to draw on.
    arrow_head_width: float
        Width of arrow head; only if show_arrows is true.
    arrow_width: float
        Width the the arrow body; only if show_arrows is true.
    arrow_cmap: float
        Cmap to color arrows; default is black arrows, but if specified will
        color the arrow by the average expression of the ligand and receptor
        of the spots connected by the arrow.
    arrow_vmax: float
        Maximum value of the arrow colour bar.
    sig_cci: bool
        Whether to only show results which involve signficant celltype-celltype
        interactions; particularly relevant when plotting the arrows.
    lr_colors: dict
        Specifies the colors of the LRs when plotting with outer_mode='binary';
        structures is {'l': color, 'r': color, 'lr': color, '': color};
        the last key-value indicates colour for spots not expressing the ligand
        or receptor.
    figsize: tuple
        (width, height) of figure if not inputted.
    kwargs:
        Extra arguments parsed to plotting functions used internally.
    """

    # Input checking #
    l, r = lr.split("_")
    ran_lr = "lr_summary" in adata.uns
    ran_sig = False if not ran_lr else "n_spots_sig" in adata.uns["lr_summary"].columns
    if ran_lr and lr in adata.uns["lr_summary"].index:
        if ran_sig:
            lr_sig = adata.uns["lr_summary"].loc[lr, :].values[1] > 0
        else:
            lr_sig = True
    else:
        lr_sig = False

    if sig_spots and not ran_lr:
        raise Exception(
            "No LR results testing results found, "
            "please run st.tl.cci_rank.run first, or set sig_spots=False."
        )

    elif sig_spots and not lr_sig:
        raise Exception(
            "LR has no significant spots, to visualise anyhow set" "sig_spots=False"
        )

    # Making sure have run_cci first with respective labelling #
    if (
        show_arrows
        and sig_cci
        and use_label
        and f"per_lr_cci_{use_label}" not in adata.uns
    ):
        raise Exception(
            "Cannot subset arrow interactions to significant ccis "
            "without performing st.tl.run_cci with "
            f"use_label={use_label} first."
        )

    # Getting which are the allowed stats for the lr to plot #
    if not ran_sig:
        lr_use_labels = ["lr_scores"]
    else:
        lr_use_labels = [
            "lr_scores",
            "p_val",
            "p_adj",
            "-log10(p_adj)",
            "lr_sig_scores",
        ]

    if type(use_mix) != type(None) and use_mix not in adata.uns:
        raise Exception(
            f"Specified use_mix, but no deconvolution results added "
            "to adata.uns matching the use_mix ({use_mix}) key."
        )
    elif (
        type(use_label) != type(None)
        and use_label in lr_use_labels
        and ran_sig
        and not lr_sig
    ):
        raise Exception(
            f"Since use_label refers to lr stats & ran permutation testing, "
            f"LR needs to be significant to view stats."
        )
    elif (
        type(use_label) != type(None)
        and use_label not in adata.obs.keys()
        and use_label not in lr_use_labels
    ):
        raise Exception(
            f"use_label must be in adata.obs or " f"one of lr stats: {lr_use_labels}."
        )

    out_options = ["binary", "continuous", None]
    if outer_mode not in out_options:
        raise Exception(f"{outer_mode} should be one of {out_options}")

    if l not in adata.var_names or r not in adata.var_names:
        raise Exception("L or R not found in adata.var_names.")

    # Whether to show just the significant spots or all spots
    lr_index = np.where(adata.uns["lr_summary"].index.values == lr)[0][0]
    sig_bool = adata.obsm["lr_sig_scores"][:, lr_index] > 0
    if sig_spots:
        adata_full = adata
        adata = adata[sig_bool, :]
    else:
        adata_full = adata

    # Dealing with the axis #
    if type(fig) == type(None) or type(ax) == type(None):
        fig, ax = plt.subplots(figsize=figsize)

    expr = adata.to_df()
    l_expr = expr.loc[:, l].values
    r_expr = expr.loc[:, r].values
    # Adding binary points of the ligand/receptor pair #
    if outer_mode == "binary":
        l_bool, r_bool = l_expr > min_expr, r_expr > min_expr
        lr_binary_labels = []
        for i in range(len(l_bool)):
            if l_bool[i] and not r_bool[i]:
                lr_binary_labels.append(l)
            elif not l_bool[i] and r_bool[i]:
                lr_binary_labels.append(r)
            elif l_bool[i] and r_bool[i]:
                lr_binary_labels.append(lr)
            elif not l_bool[i] and not r_bool[i]:
                lr_binary_labels.append("")
        lr_binary_labels = pd.Series(
            np.array(lr_binary_labels), index=adata.obs_names
        ).astype("category")
        adata.obs[f"{lr}_binary_labels"] = lr_binary_labels

        if type(lr_cmap) == type(None):
            lr_cmap = "default"  # This gets ignored due to setting colours below
            if type(lr_colors) == type(None):
                lr_colors = {
                    l: matplotlib.colors.to_hex("r"),
                    r: matplotlib.colors.to_hex("limegreen"),
                    lr: matplotlib.colors.to_hex("b"),
                    "": "#836BC6",  # Neutral color in H&E images.
                }

            label_set = adata.obs[f"{lr}_binary_labels"].cat.categories
            adata.uns[f"{lr}_binary_labels_colors"] = [
                lr_colors[label] for label in label_set
            ]
        else:
            lr_cmap = check_cmap(lr_cmap)

        cluster_plot(
            adata,
            use_label=f"{lr}_binary_labels",
            cmap=lr_cmap,
            size=outer_size_prop * pt_scale,
            crop=False,
            ax=ax,
            fig=fig,
            show_image=show_image,
            show_plot=False,
            **kwargs,
        )

    # Showing continuous gene expression of the LR pair #
    elif outer_mode == "continuous":
        if type(l_cmap) == type(None):
            l_cmap = matplotlib.colors.LinearSegmentedColormap.from_list(
                "lcmap", [(0, 0, 0), (0.5, 0, 0), (0.75, 0, 0), (1, 0, 0)]
            )
        else:
            l_cmap = check_cmap(l_cmap)
        if type(r_cmap) == type(None):
            r_cmap = matplotlib.colors.LinearSegmentedColormap.from_list(
                "rcmap", [(0, 0, 0), (0, 0.5, 0), (0, 0.75, 0), (0, 1, 0)]
            )
        else:
            r_cmap = check_cmap(r_cmap)

        gene_plot(
            adata,
            gene_symbols=l,
            size=outer_size_prop * pt_scale,
            cmap=l_cmap,
            color_bar_label=l,
            ax=ax,
            fig=fig,
            crop=False,
            show_image=show_image,
            **kwargs,
        )
        gene_plot(
            adata,
            gene_symbols=r,
            size=middle_size_prop * pt_scale,
            cmap=r_cmap,
            color_bar_label=r,
            ax=ax,
            fig=fig,
            crop=False,
            show_image=show_image,
            **kwargs,
        )

    # Adding the cell type labels #
    if type(use_label) != type(None):
        if use_label in lr_use_labels:
            inner_cmap = inner_cmap if type(inner_cmap) != type(None) else "copper"
            # adata.obsm[f'{lr}_{use_label}'] = adata.uns['per_lr_results'][
            #                          lr].loc[adata.obs_names,use_label].values
            lr_result_plot(
                adata,
                use_lr=lr,
                show_image=show_image,
                cmap=inner_cmap,
                crop=False,
                ax=ax,
                fig=fig,
                size=inner_size_prop * pt_scale,
                **kwargs,
            )
        else:
            inner_cmap = inner_cmap if type(inner_cmap) != type(None) else "default"
            cluster_plot(
                adata,
                use_label=use_label,
                cmap=inner_cmap,
                size=inner_size_prop * pt_scale,
                crop=False,
                ax=ax,
                fig=fig,
                show_image=show_image,
                show_plot=False,
                **kwargs,
            )

    # Adding in labels which show the interactions between signicant spots &
    # neighbours
    if show_arrows:
        l_expr = adata_full[:, l].X.toarray()[:, 0]
        r_expr = adata_full[:, r].X.toarray()[:, 0]

        if sig_cci:
            int_df = adata.uns[f"per_lr_cci_{use_label}"][lr]
        else:
            int_df = None

        cci_hs.add_arrows(
            adata_full,
            l_expr,
            r_expr,
            min_expr,
            sig_bool,
            fig,
            ax,
            use_label,
            int_df,
            arrow_head_width,
            arrow_width,
            arrow_cmap,
            arrow_vmax,
        )

    # Cropping #
    # if crop:
    #     x0, x1 = ax.get_xlim()
    #     y0, y1 = ax.get_ylim()
    #     x_margin, y_margin = (x1-x0)*margin_ratio, (y1-y0)*margin_ratio
    #     print(x_margin, y_margin)
    #     print(x0, x1, y0, y1)
    #     ax.set_xlim(x0 - x_margin, x1 + x_margin)
    #     ax.set_ylim(y0 - y_margin, y1 + y_margin)
    #     #ax.set_ylim(ax.get_ylim()[::-1])

    fig.suptitle(title)


#### het_plot currently out of date;
#### from old data structure when only test individual LRs.
@_docs_params(spatial_base_plot=doc_spatial_base_plot, het_plot=doc_het_plot)
def het_plot(
    adata: AnnData,
    # plotting param
    title: Optional["str"] = None,
    figsize: Optional[Tuple[float, float]] = None,
    cmap: Optional[str] = "Spectral_r",
    use_label: Optional[str] = None,
    list_clusters: Optional[list] = None,
    ax: Optional[matplotlib.axes._subplots.Axes] = None,
    fig: Optional[matplotlib.figure.Figure] = None,
    show_plot: Optional[bool] = True,
    show_axis: Optional[bool] = False,
    show_image: Optional[bool] = True,
    show_color_bar: Optional[bool] = True,
    zoom_coord: Optional[float] = None,
    crop: Optional[bool] = True,
    margin: Optional[bool] = 100,
    size: Optional[float] = 7,
    image_alpha: Optional[float] = 1.0,
    cell_alpha: Optional[float] = 1.0,
    use_raw: Optional[bool] = False,
    fname: Optional[str] = None,
    dpi: Optional[int] = 120,
    # cci_rank param
    use_het: Optional[str] = "het",
    contour: bool = False,
    step_size: Optional[int] = None,
    vmin: float = None,
    vmax: float = None,
) -> Optional[AnnData]:
    """\
    Allows the visualization of significant cell-cell interaction
    as the values of dot points or contour in the Spatial
    transcriptomics array.


    Parameters
    -------------------------------------
    {spatial_base_plot}
    {het_plot}

    Examples
    -------------------------------------
    >>> import stlearn as st
    >>> adata = st.datasets.example_bcba()
    >>> pvalues = "lr_pvalues"
    >>> st.pl.gene_plot(adata, use_het = pvalues)

    """

    CciPlot(
        adata,
        title=title,
        figsize=figsize,
        cmap=cmap,
        use_label=use_label,
        list_clusters=list_clusters,
        ax=ax,
        fig=fig,
        show_plot=show_plot,
        show_axis=show_axis,
        show_image=show_image,
        show_color_bar=show_color_bar,
        zoom_coord=zoom_coord,
        crop=crop,
        margin=margin,
        size=size,
        image_alpha=image_alpha,
        cell_alpha=cell_alpha,
        use_raw=use_raw,
        fname=fname,
        dpi=dpi,
        use_het=use_het,
        contour=contour,
        step_size=step_size,
        vmin=vmin,
        vmax=vmax,
    )


# Functions relating to visualising celltype-celltype interactions after
# calling: st.tl.cci.run_cci


def ccinet_plot(
    adata: AnnData,
    use_label: str,
    lr: str = None,
    pos: dict = None,
    return_pos: bool = False,
    cmap: str = "default",
    font_size: int = 12,
    node_size_exp: int = 1,
    node_size_scaler: int = 1,
    min_counts: int = 0,
    sig_interactions: bool = True,
    fig: matplotlib.figure.Figure = None,
    ax: matplotlib.axes.Axes = None,
    pad=0.25,
    title: str = None,
    figsize: tuple = (10, 10),
):
    """Circular celltype-celltype interaction network based on LR-CCI analysis.
    The size of the nodes drawn for each cell type indicates the total no. of
    spot interactions that cell type is involved in; while the color of the
    arrows between nodes is coloured by the total no. of interactions between
    those particular cell types.

    Parameters
    ----------
    adata: AnnData
        Data on which st.tl.cci.run & st.tl.cci.run_cci has been applied.
    use_label: str
        Indicates the cell type labels or deconvolution results used for
        cell-cell interaction counting by LR pairs.
    lr: str
        The LR pair to visualise the cci network for. If None, will use spot
        cci counts across all LR pairs from adata.uns[f'lr_cci_{use_label}'].
    pos: dict
        Positions to draw each cell type, format as outputted from running
        networkx.circular_layout(graph). If not inputted will be generated.
    return_pos: bool
        Whether to return the positions of the cell types drawn or not;
        useful for input back into this function via the 'pos' parameter to get
        consistent positioning of the cell types when plotting for different LR
        pairs.
    cmap: str
        Cmap to use when generating the cell colors, if not already specified by
        adata.uns[f'{use_label}_colors'].
    font_size: int
        Size of the cell type labels.
    node_size_scaler: float
        Scaler to multiply by node sizes to increase/decrease size.
    node_size_exp: int
        Increases difference between node sizes by this exponent.
    min_counts: int
        Minimum no. of LR interactions for connection to be drawn.

    Returns
    -------
    pos: dict
        Dictionary of positions where the nodes are draw if return_pos is True, useful for consistent layouts.
    """
    cmap, cmap_n = get_cmap(cmap)
    # Making sure adata in correct state that this function should run #
    if f"lr_cci_{use_label}" not in adata.uns:
        raise Exception(
            "Need to first call st.tl.run_cci with the equivalnt "
            "use_label to visualise cell-cell interactions."
        )
    elif type(lr) != type(None) and lr not in adata.uns[f"per_lr_cci_{use_label}"]:
        raise Exception(
            f"{lr} not found in {f'per_lr_cci_{use_label}'}, "
            "suggesting no significant interactions."
        )

    # Either plotting overall interactions, or just for a particular LR #
    int_df, title = get_int_df(adata, lr, use_label, sig_interactions, title)

    # Creating the interaction graph #
    all_set = int_df.index.values
    int_matrix = int_df.values
    graph = nx.MultiDiGraph()
    int_bool = int_matrix > min_counts
    int_matrix = int_matrix * int_bool
    for i, cell_A in enumerate(all_set):
        if cell_A not in graph:
            graph.add_node(cell_A)
        for j, cell_B in enumerate(all_set):
            if int_bool[i, j]:
                count = int_matrix[i, j]
                graph.add_edge(cell_A, cell_B, weight=count)

    # Determining graph layout, node sizes, & edge colours #
    if type(pos) == type(None):
        pos = nx.circular_layout(graph)  # position the nodes using the layout
    total = sum(sum(int_matrix))
    node_names = list(graph.nodes.keys())
    node_indices = [np.where(all_set == node_name)[0][0] for node_name in node_names]
    node_sizes = np.array(
        [
            (
                ((sum(int_matrix[i, :] + int_matrix[:, i]) - int_matrix[i, i]) / total)
                * 10000
                * node_size_scaler
            )
            ** (node_size_exp)
            for i in node_indices
        ]
    )
    node_sizes[node_sizes == 0] = 0.1  # pseudocount

    edges = list(graph.edges.items())
    e_totals = []
    for i, edge in enumerate(edges):
        trans_i = np.where(all_set == edge[0][0])[0][0]
        receive_i = np.where(all_set == edge[0][1])[0][0]
        e_total = (
            sum(list(int_matrix[trans_i, :]) + list(int_matrix[:, receive_i]))
            - int_matrix[trans_i, receive_i]
        )  # so don't double count
        e_totals.append(e_total)
    edge_weights = [edge[1]["weight"] / e_totals[i] for i, edge in enumerate(edges)]

    # Determining node colors #
    nodes = np.unique(list(graph.nodes.keys()))
    node_colors = get_colors(adata, use_label, cmap, label_set=nodes)
    if not np.all(np.array(node_names) == nodes):
        nodes_indices = [np.where(nodes == node)[0][0] for node in node_names]
        node_colors = np.array(node_colors)[nodes_indices]

    #### Drawing the graph #####
    if type(fig) == type(None) or type(ax) == type(None):
        fig, ax = plt.subplots(figsize=figsize, facecolor=[0.7, 0.7, 0.7, 0.4])

    # Adding in the self-loops #
    z = 55
    for i, edge in enumerate(edges):
        cell_type = edge[0][0]
        if cell_type != edge[0][1]:
            continue
        x, y = pos[cell_type]
        angle = math.degrees(math.atan(y / x))
        if x > 0:
            angle = angle + 180
        arc = patches.Arc(
            xy=(x, y),
            width=0.3,
            height=0.025,
            lw=5,
            ec=plt.cm.get_cmap("Blues")(edge_weights[i]),
            angle=angle,
            theta1=z,
            theta2=360 - z,
        )
        ax.add_patch(arc)

    # Drawing the main components of the graph #
    edges = nx.draw_networkx(
        graph,
        pos,
        node_size=node_sizes,
        node_color=node_colors,
        arrowstyle="->",
        arrowsize=50,
        width=5,
        font_size=font_size,
        font_weight="bold",
        edge_color=edge_weights,
        edge_cmap=plt.cm.Blues,
        ax=ax,
    )
    fig.suptitle(title, fontsize=30)
    plt.tight_layout()

    # Adding padding #
    xlims = ax.get_xlim()
    ax.set_xlim(xlims[0] - pad, xlims[1] + pad)
    ylims = ax.get_ylim()
    ax.set_ylim(ylims[0] - pad, ylims[1] + pad)

    if return_pos:
        return pos


def cci_map(
    adata: AnnData,
    use_label: str,
    lr: str = None,
    ax: matplotlib.figure.Axes = None,
    show: bool = False,
    figsize: tuple = None,
    cmap: str = "Spectral_r",
    sig_interactions: bool = True,
    title=None,
):
    """Heatmap visualising sender->receivers of cell type interactions.

    Parameters
    ----------
    adata: AnnData
        Data on which st.tl.cci.run & st.tl.cci.run_cci has been applied.
    use_label: str
        Indicates the cell type labels or deconvolution results used for
        cell-cell interaction counting by LR pairs.
    lr: str
        The LR pair to visualise the sender->receiver interactions for.
        If None, will use all pairs via adata.uns[f'lr_cci_{use_label}'].
    ax: Axes
        Axes on which to plot the heatmap, if None then generates own.
    show: bool
        Whether to show the plot or not; if not, then returns ax.
    figsize: tuple
        (width, height), specifies the dimensions of the figure. Only relevant
        if ax=None.
    cmap: str
        Cmap used to color the number of LR interactions.
    sig_interactions: bool
        Whether to only show significant CCIs, or all observed interactions.
    title: None
        Title to display over the heatmap. If not provided, will be determined
        based on the run parameters.

    Returns
    -------
    ax: matplotlib.figure.Axes
        Axes where the heatmap was drawn if show=False.
    """

    # Either plotting overall interactions, or just for a particular LR #
    int_df, title = get_int_df(adata, lr, use_label, sig_interactions, title)

    if type(figsize) == type(None):  # Adjust size depending on no. cell types
        add = np.array([int_df.shape[0] * 0.1, int_df.shape[0] * 0.05])
        figsize = tuple(np.array([6.4, 4.8]) + add)

    # Rank by total interactions #
    int_vals = int_df.values
    total_ints = int_vals.sum(axis=1) + int_vals.sum(axis=0) - int_vals.diagonal()
    order = np.argsort(-total_ints)
    int_df = int_df.iloc[order, order[::-1]]

    # Reformat the interaction df #
    flat_df = create_flat_df(int_df)

    ax = _box_map(
        flat_df["x"],
        flat_df["y"],
        flat_df["value"].astype(int),
        ax=ax,
        figsize=figsize,
        cmap=cmap,
    )

    ax.set_ylabel("Sender")
    ax.set_xlabel("Receiver")
    plt.suptitle(title)

    if show:
        plt.show()
    else:
        return ax


def lr_cci_map(
    adata: AnnData,
    use_label: str,
    lrs: list or np.array = None,
    n_top_lrs: int = 5,
    n_top_ccis: int = 15,
    min_total: int = 0,
    ax: matplotlib.figure.Axes = None,
    figsize: tuple = (6.48, 4.8),
    show: bool = False,
    cmap: str = "Spectral_r",
    square_scaler: int = 700,
    sig_interactions: bool = True,
):
    """Heatmap of interaction counts.
        Rows are lrs and columns are celltype->celltype interactions.

    Parameters
    ----------
    adata: AnnData
        Data on which st.tl.cci.run & st.tl.cci.run_cci has been applied.
    use_label: str
        Indicates the cell type labels or deconvolution results used for
        the cell-cell interaction counting by LR pairs.
    lrs: list-like
        LR pairs to show in the heatmap, if None then top 5 lrs with highest no.
        of interactions used from adata.uns['lr_summary'].
    n_top_lrs: int
        Indicates how many top lrs to show; is ignored if lrs is not None.
    n_top_ccis: int
        Indicates maximum no. of CCIs to show.
    min_total: int
        Minimum no. of totals interaction celltypes must have to be shown.
    ax: Axes
        Axes on which to draw the heatmap, is generated internally if None.
    figsize: tuple
        (width, height), only relevant if ax=None.
    show: bool
        Whether to show the plot or not, if not returns ax.
    cmap: str
        Cmap used to color the number of LR interactions.
    square_scaler: int
        Scaler to size the squares displayed.
    sig_interactions: bool
        Whether to only show significant CCIs, or all observed interactions.

    Returns
    -------
    ax: matplotlib.figure.Axes
        Axes where the heatmap was drawn on if show=False.
    """
    if sig_interactions:
        lr_int_dfs = adata.uns[f"per_lr_cci_{use_label}"]
    else:
        lr_int_dfs = adata.uns[f"per_lr_cci_raw_{use_label}"]

    if type(lrs) == type(None):
        lrs = np.array(list(lr_int_dfs.keys()))
    else:
        lrs = np.array(lrs)
        n_top_lrs = len(lrs)

    # Creating a new int_df with lrs as rows & cell-cell as column #
    cell_types = list(lr_int_dfs.values())[0].index.values.astype(str)
    n_ints = len(cell_types) ** 2
    new_ints = np.zeros((len(lrs), n_ints))
    for lr_i, lr in enumerate(lrs):
        col_i = 0
        int_df = lr_int_dfs[lr]
        ccis = []
        for c_i, cell_i in enumerate(cell_types):
            for c_j, cell_j in enumerate(cell_types):
                new_ints[lr_i, col_i] = int_df.values[c_i, c_j]
                ccis.append("->".join([cell_i, cell_j]))
                col_i += 1
    new_int_df = pd.DataFrame(new_ints, index=lrs, columns=ccis)

    # Filtering out ccis which have few LR interactions #
    total_ints = new_int_df.values.sum(axis=0)
    order = np.argsort(-total_ints)
    new_int_df = new_int_df.iloc[:, order[0:n_top_ccis]]

    # Getting the top_lrs to display by top loadings in PCA #
    if n_top_lrs < len(lrs):
        top_lrs = adata.uns["lr_summary"].index.values[0:n_top_lrs]
        new_int_df = new_int_df.loc[top_lrs, :]

    # Ordering by the no. of interactions #
    cci_ints = new_int_df.values.sum(axis=0)
    cci_order = np.argsort(-cci_ints)
    lr_ints = new_int_df.values.sum(axis=1)
    lr_order = np.argsort(-lr_ints)
    new_int_df = new_int_df.iloc[lr_order, cci_order]

    # Getting a flat version of the array for plotting #
    flat_df = create_flat_df(new_int_df.transpose())
    if flat_df.shape[0] == 0 or flat_df.shape[1] == 0:
        raise Exception(f"No interactions greater than min: {min_total}")

    ax = _box_map(
        flat_df["x"],
        flat_df["y"],
        flat_df["value"].astype(int),
        ax=ax,
        cmap=cmap,
        figsize=figsize,
        square_scaler=square_scaler,
    )

    ax.set_ylabel("LR-pair")
    ax.set_xlabel("Cell-cell interaction")

    if show:
        plt.show()
    else:
        return ax


def lr_chord_plot(
    adata: AnnData,
    use_label: str,
    lr: str = None,
    min_ints: int = 2,
    n_top_ccis: int = 10,
    cmap: str = "default",
    sig_interactions: bool = True,
    label_size: int = 10,
    label_rotation: float = 0,
    title: str = None,
    figsize: tuple = (8, 8),
    show: bool = True,
):
    """Chord diagram of interactions between cell types.
        Note that interaction is measured as the total no. of edges connecting
        two cell types expressing the ligand and/or receptor in significant
        neighbourhoods for given LR pair.

        The chord diagram is read as follows:

        Each cell type has a labelled edge taking up a proportion of the outter circle.
        Chords connecting cell type edges are coloured by the dominant sending cell.
        Each chord linking cell types has an assymetric shape.
        For two cell types, A and B, the side of the chord attached to edge A is sized by
        the total interactions from B->A, where B is expressing the ligand & A
        is expressing the receptor.
        Hence, the proportion of a cell type's edge in the chordplot circle
        represents the total input signals to that cell type; while the
        area of the chordplot circle taken up by the outputted chords from a given
        cell type represents the total output signals from that cell type.

    Parameters
    ----------
    adata: AnnData
        Data on which st.tl.cci.run & st.tl.cci.run_cci has been applied.
    use_label: str
        Indicates the cell type labels or deconvolution results used for
        cell-cell interaction counting by LR pairs.
    lr: str
        The LR pair to visualise the CCIs for.
        If None, will use all pairs via adata.uns[f'lr_cci_{use_label}'].
    min_ints: int
        Minimum no. of interactions celltypes must have to be shown.
    n_top_ccis: int
        Maximum no. of CCIs to show, will take the top number of these to display.
    cmap: str
        Cmap to use to get colors if colors not already in adata.uns[f'{use_label}_colors']
    sig_interactions: bool
        Whether to show only significant CCIs or all interaction counts.
    label_size: str
        The size of the cell type labels to render.
    label_rotation: float
        Rotation of the cell type label text.
    title: str
        The title above the plot; informative default is determined based on input.
    figsize: tuple
        Figure dimensions.
    show: bool
        Show or not; if not return figure & axes.
    Returns
    -------
    fig: matplotlib.figure.Figure
        Figure on which the heatmap was drawn if show=False.
    ax: matplotlib.figure.Axes
        Axes where the heatmap was drawn on if show=False.
    """
    # Either plotting overall interactions, or just for a particular LR #
    int_df, title = get_int_df(adata, lr, use_label, sig_interactions, title)

    int_df = int_df.transpose()
    fig = plt.figure(figsize=figsize)

    flux = int_df.values
    total_ints = flux.sum(axis=1) + flux.sum(axis=0) - flux.diagonal()
    keep = np.where(total_ints > min_ints)[0]
    # Limit of 10 for good display #
    if len(keep) > n_top_ccis:
        keep = np.argsort(-total_ints)[0:n_top_ccis]
    # Filter any with all zeros after filtering #
    all_zero = np.array(
        [np.all(np.logical_and(flux[i, keep] == 0, flux[keep, i] == 0)) for i in keep]
    )
    keep = keep[all_zero == False]
    if len(keep) == 0:  # If we don't keep anything, warn the user
        print(
            f"Warning: for {lr} at the current min_ints ({min_ints}), there "
            f"are no interaction to display. Adjust min_ints to a lower value"
            f" to visualise chordplot for this LR."
        )
        return

    flux = flux[:, keep]
    flux = flux[keep, :].astype(float)

    # Add pseudocount to row/column which has all zeros for the incoming
    # so can make the connection between the two
    for i in range(flux.shape[0]):
        if np.all(flux[i, :] == 0):
            flux[i, flux[:, i] > 0] += sys.float_info.min
        elif np.all(flux[:, i] == 0):
            flux[flux[i, :] > 0, i] += sys.float_info.min

    cell_names = int_df.index.values.astype(str)[keep]
    nodes = cell_names

    # Retrieving colors of cell types #
    colors = get_colors(adata, use_label, cmap=cmap, label_set=cell_names)

    ax = plt.axes([0, 0, 1, 1])
    nodePos = chordDiagram(flux, ax, lim=1.25, colors=colors)
    ax.axis("off")
    prop = dict(fontsize=label_size, ha="center", va="center")
    label_rotation_ = label_rotation
    for i in range(len(cell_names)):
        x, y = nodePos[i][0:2]
        rotation = nodePos[i][2]
        # Prevent text going upside down at certain rotations
        if (rotation < 90 and rotation > 18 and label_rotation != 0) or (
            rotation < 120 and rotation > 90
        ):
            label_rotation_ = -label_rotation
        else:
            label_rotation_ = label_rotation
        ax.text(
            x, y, nodes[i], rotation=nodePos[i][2] + label_rotation_, **prop
        )  # size=10,
    fig.suptitle(title, fontsize=12, fontweight="bold")
    if show:
        plt.show()
    else:
        return fig, ax


def grid_plot(
    adata,
    use_label: str = None,
    n_row: int = 10,
    n_col: int = 10,
    size: int = 1,
    figsize=(4.5, 4.5),
    show: bool = False,
):
    """Plots grid over the top of spatial data to show how cells will be grouped if gridded.

    Parameters
    ----------
    adata: AnnData
        Data on which st.tl.cci.run & st.tl.cci.run_cci has been applied.
    use_label: str
        Indicates the cell type labels or deconvolution results used for
        cell-cell interaction counting by LR pairs.
    n_row: str
        The number of rows in the grid.
    n_col: int
        The number of columns in the grid.

    Returns
    -------
    fig: matplotlib.figure.Figure
        Figure on which the heatmap is draw if show=False.
    ax: matplotlib.figure.Axes
        Axes where the heatmap was drawn on if show=False.
    """
    xs, ys = adata.obs["imagecol"].values, adata.obs["imagerow"].values
    grid_counts, xedges, yedges = np.histogram2d(xs, ys, bins=[n_col, n_row])
    xmin, xmax = min(xedges), max(xedges)
    ymin, ymax = min(yedges), max(yedges)

    fig, ax = plt.subplots(figsize=figsize)

    # Plotting the points #
    if type(use_label) != type(None):
        if f"{use_label}_colors" in adata.uns:
            color_map = {}
            for i, ct in enumerate(adata.obs[use_label].cat.categories):
                color_map[ct] = adata.uns[f"{use_label}_colors"][i]
            cell_colors = [color_map[ct] for ct in adata.obs[use_label]]
    else:  # Otherwise plot by cell density #
        stack = np.vstack([xs, ys])
        cell_colors = gaussian_kde(stack)(stack)

    ax.scatter(xs, -ys, s=size, c=cell_colors)
    ax.vlines(xedges, -ymin, -ymax, color="#36454F")
    ax.hlines(-yedges, xmin, xmax, color="#36454F")

    if show:
        plt.show()
    else:
        return fig, ax


####################### Bokeh Interactive Plots ################################
def lr_plot_interactive(adata: AnnData):
    """Plots the LR scores for significant spots interatively using Bokeh.

    Parameters
    ----------
    adata: AnnData
        Data on which st.tl.cci.run has been applied.
    """
    bokeh_object = BokehLRPlot(adata)
    output_notebook()
    show(bokeh_object.app, notebook_handle=True)


def spatialcci_plot_interactive(adata: AnnData):
    """Plots the significant CCI in the spatial context interactively using Bokeh.

    Parameters
    ----------
    adata: AnnData
        Data on which st.tl.cci.run & st.tl.cci.run_cci has been applied.
    """
    bokeh_object = BokehSpatialCciPlot(adata)
    output_notebook()
    show(bokeh_object.app, notebook_handle=True)


# def het_plot_interactive(adata: AnnData):
#     bokeh_object = BokehCciPlot(adata)
#     output_notebook()
#     show(bokeh_object.app, notebook_handle=True)


# Bokeh & old grid plots;
# has not been tested since multi-LR testing implimentation.

# def het_plot_interactive(adata: AnnData):
#     bokeh_object = BokehCciPlot(adata)
#     output_notebook()
#     show(bokeh_object.app, notebook_handle=True)


# def grid_plot(
# 	adata: AnnData,
# 	use_het: str = None,
# 	num_row: int = 10,
# 	num_col: int = 10,
# 	vmin: float = None,
# 	vmax: float = None,
# 	cropped: bool = True,
# 	margin: int = 100,
# 	dpi: int = 100,
# 	name: str = None,
# 	output: str = None,
# 	copy: bool = False,
# ) -> Optional[AnnData]:
#
# 	"""
# 	Cell diversity plot for sptial transcriptomics data.
#
# 	Parameters
# 	----------
# 	adata:                  Annotated data matrix.
# 	use_het:                Cluster heterogeneity count results from tl.cci_rank.het
# 	num_row: int            Number of grids on height
# 	num_col: int            Number of grids on width
# 	cropped                 crop image or not.
# 	margin                  margin used in cropping.
# 	dpi:                    Set dpi as the resolution for the plot.
# 	name:                   Name of the output figure file.
# 	output:                 Save the figure as file or not.
# 	copy:                   Return a copy instead of writing to adata.
#
# 	Returns
# 	-------
# 	Nothing
# 	"""
#
# 	try:
# 		import seaborn as sns
# 	except:
# 		raise ImportError("Please run `pip install seaborn`")
# 	plt.subplots()
#
# 	sns.heatmap(
# 		pd.DataFrame(np.array(adata.obsm[use_het]).reshape(num_col, num_row)).T,
# 		vmin=vmin,
# 		vmax=vmax,
# 	)
# 	plt.axis("equal")
#
# 	if output is not None:
# 		plt.savefig(
# 			output + "/" + name + "_heatmap.pdf",
# 			dpi=dpi,
# 			bbox_inches="tight",
# 			pad_inches=0,
# 		)
#
# 	plt.show()



""" Performs LR analysis by grouping LR pairs which having hotspots across
    similar tissues.
"""

from stlearn.pl import het_plot
from sklearn.cluster import DBSCAN, AgglomerativeClustering
from anndata import AnnData
from tqdm import tqdm
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sb


def get_hotspots(
    adata: AnnData,
    lr_scores: np.ndarray,
    lrs: np.array,
    eps: float,
    quantile=0.05,
    verbose=True,
    plot_diagnostics: bool = False,
    show_plot: bool = False,
):
    """Determines the hotspots for the inputted scores by progressively setting more stringent cutoffs & cluster in space, chooses point which maximises number of clusters.
    Parameters
    ----------
    adata: AnnData          The data object
    lr_scores: np.ndarray      LR_pair*Spots containing the LR scores.
    lrs: np.array           The LR_pairs, in-line with the rows of scores.
    eps: float              The eps parameter used in DBScan to get the number of clusters.
    quantile: float         The quantiles to use for the cutoffs, if 0.05 then will take non-zero quantiles of 0.05, 0.1,..., 1 quantiles to cluster.

    Returns
    -------
    lr_hot_scores: np.ndarray, lr_cutoffs: np.array          First is the LR scores for just the hotspots, second is the cutoff used to get those LR_scores.
    """
    coors = adata.obs[["imagerow", "imagecol"]].values
    lr_summary, lr_hot_scores = hotspot_core(
        lr_scores, lrs, coors, eps, quantile, plot_diagnostics, adata
    )

    if plot_diagnostics and show_plot:  # Showing the diagnostic plotting #
        plt.show()

    if verbose:
        print("Clustering LRs to help with ranking/interpretation...")
    # Clustering the LR pairs to obtain a set of clusters so to order within
    # each cluster
    clusterer = AgglomerativeClustering(
        affinity="euclidean", linkage="ward", distance_threshold=10, n_clusters=None
    )
    clusterer.fit(lr_hot_scores > 0)
    dist_cutoff = np.quantile(clusterer.distances_, 0.98)
    clusterer = AgglomerativeClustering(
        affinity="euclidean",
        linkage="ward",
        distance_threshold=dist_cutoff,
        n_clusters=None,
    )
    clusters = clusterer.fit_predict(lr_hot_scores > 0)
    cluster_set = np.unique(clusters)

    if verbose:
        print("Ranking LRs...")

    # Determining the ordering of the clusters so is useful to user #
    cluster_mean_spots = []
    for cluster in cluster_set:
        cluster_bool = clusters == cluster
        cluster_mean_spots.append(np.mean(lr_summary[cluster_bool, 2]))
    cluster_order = np.argsort(-np.array(cluster_mean_spots))

    # Determining order of lrs in cluster & also overall cluster scores #
    lr_order = []
    new_clusters = []
    cluster_scores = np.zeros((adata.shape[0], len(cluster_set)))
    for i, index in enumerate(cluster_order):
        cluster = cluster_set[index]
        cluster_indices = np.where(clusters == cluster)[0]
        lr_order_ = np.argsort(-lr_summary[cluster_indices, 2])
        lr_order.extend(cluster_indices[lr_order_])

        new_clusters += [i] * len(cluster_indices)

        cluster_scores[:, i] = lr_hot_scores[cluster_indices, :].mean(axis=0)

    if verbose:
        print("Saving results:")

    # Re-ordering the summary & the scores #
    lrs = lrs[lr_order]
    lr_summary = lr_summary[lr_order, :]
    lr_summary[:, 3] = new_clusters
    lr_summary = pd.DataFrame(
        lr_summary,
        index=lrs,
        columns=["spot_counts", "cutoff", "hotspot_counts", "lr_cluster"],
    )
    lr_scores = lr_scores[lr_order, :].transpose()
    lr_hot_scores = lr_hot_scores[lr_order, :].transpose()

    # Adding all this information to the AnnData #
    adata.uns["lr_summary"] = lr_summary
    adata.obsm["lr_scores"] = lr_scores
    adata.obsm["lr_hot_scores"] = lr_hot_scores
    adata.obsm["cluster_scores"] = cluster_scores

    if verbose:
        print(f"\tSummary values of lrs in adata.uns['lr_summary'].")
        print(
            f"\tMatrix of lr scores in same order as the summary in adata.obsm['lr_scores']."
        )
        print(f"\tMatrix of the hotspot scores in adata.obsm['lr_hot_scores'].")
        print(
            f"\tMatrix of the mean LR cluster scores in adata.obsm['cluster_scores']."
        )


def hotspot_core(
    lr_scores,
    lrs,
    coors,
    eps,
    quantile,
    plot_diagnostics=False,
    adata=None,
    verbose=True,
    max_score=False,
):
    """Made code for getting the hotspot information."""
    score_copy = lr_scores.copy()
    quantiles = [quantile * i for i in range(int(1 / quantile))]

    # Values to return #
    lr_hot_scores = np.zeros(score_copy.shape)
    # cols: spot_counts, cutoff, hotspot_counts, lr_cluster
    lr_summary = np.zeros((score_copy.shape[0], 4))

    ### Also creating grouping lr_pairs by quantiles to plot diagnostics ###
    if plot_diagnostics:
        lr_quantiles = [(i / 6) for i in range(1, 7)][::-1]
        lr_mean_scores = np.apply_along_axis(non_zero_mean, 1, score_copy)
        lr_quant_values = np.quantile(lr_mean_scores, lr_quantiles)
        quant_lrs = np.array(
            [lrs[lr_mean_scores == quant] for quant in lr_quant_values]
        )
        fig, axes = plt.subplots(6, 4, figsize=(20, 15))

    # Determining the cutoffs for hotspots #
    with tqdm(
        total=len(lrs),
        desc="Removing background lr scores...",
        bar_format="{l_bar}{bar}",
        disable=verbose == False,
    ) as pbar:
        for i, lr_ in enumerate(lrs):
            lr_score_ = score_copy[i, :]
            lr_summary[i, 0] = len(np.where(lr_score_ > 0)[0])

            cutoff_scores = []
            cutoffs = np.quantile(lr_score_[lr_score_ > 0], quantiles)
            for cutoff in cutoffs:
                spot_bool = lr_score_ >= cutoff
                if len(np.where(spot_bool)[0]) == 0:
                    cutoff_scores.append(0)
                    continue

                coor_ = coors[spot_bool, :]
                clusters = DBSCAN(
                    min_samples=2, eps=eps, metric="manhattan"
                ).fit_predict(coor_)
                score = len(np.unique(clusters)) * (np.mean(lr_score_[spot_bool])) ** 2
                cutoff_scores.append(score)

            # Cutoff point where maximum number of clusters occurs #
            best_cutoff = cutoffs[np.argmax(cutoff_scores)]
            if not max_score:
                lr_summary[i, 1] = best_cutoff
            else:
                lr_summary[i, 1] = cutoff_scores[np.argmax(cutoff_scores)]

            lr_score_[lr_score_ < best_cutoff] = 0
            lr_hot_scores[i, :] = lr_score_
            lr_summary[i, 2] = len(np.where(lr_score_ > 0)[0])

            # Adding the diagnostic plots #
            if plot_diagnostics and lr_ in quant_lrs and type(adata) != type(None):
                add_diagnostic_plots(
                    adata,
                    i,
                    lr_,
                    quant_lrs,
                    lr_quantiles,
                    lr_scores,
                    lr_hot_scores,
                    axes,
                    cutoffs,
                    cutoff_scores,
                    best_cutoff,
                )

            pbar.update(1)

    return lr_summary, lr_hot_scores


def non_zero_mean(vals):
    """Gives the non-zero mean of the values."""
    return vals[vals > 0].mean()


def add_diagnostic_plots(
    adata,
    i,
    lr_,
    quant_lrs,
    lr_quantiles,
    lr_scores,
    lr_hot_scores,
    axes,
    cutoffs,
    n_clusters,
    best_cutoff,
):
    """Adds diagnostic plots for the quantile LR pair to a figure to illustrate \
        how the cutoff is functioning.
    """
    q_i = np.where(quant_lrs == lr_)[0][0]

    # Scatter plot #
    axes[q_i][0].scatter(cutoffs, n_clusters)
    axes[q_i][0].set_title(f"n_clusts*mean_spot_score vs cutoff")
    axes[q_i][0].set_xlabel("cutoffs")
    axes[q_i][0].set_ylabel("n_clusts*mean_spot_score")

    # Distribution of scores with cutoff #
    scores_ = lr_scores[i, :]
    sb.distplot(
        scores_[scores_ > 0],
        ax=axes[q_i][1],
        hist=True,
        kde=False,
        color="red",
        norm_hist=True,
    )
    v_height = 0.5
    axes[q_i][1].vlines(best_cutoff, 0, v_height)
    axes[q_i][1].text(best_cutoff, v_height, str(round(best_cutoff, 2)))
    axes[q_i][1].set_title(f"Distrib {round(lr_quantiles[q_i], 2)}({lr_})")

    # Showing before & after filtering spots #
    adata.obsm["lr_scores"] = scores_
    het_plot(
        adata,
        use_het="lr_scores",
        ax=axes[q_i][2],
        show_color_bar=False,
    )
    axes[q_i][2].set_title("scores")

    adata.obsm["lr_scores"] = lr_hot_scores[i, :]
    het_plot(
        adata,
        use_het="lr_scores",
        ax=axes[q_i][3],
        show_color_bar=False,
    )
    axes[q_i][3].set_title("hotspot scores")



from anndata import AnnData
from typing import Optional, Union
import numpy as np
import pandas as pd
import networkx as nx
from scipy.spatial.distance import cdist
import scanpy


def pseudotime(
    adata: AnnData,
    use_label: str = None,
    eps: float = 20,
    n_neighbors: int = 25,
    use_rep: str = "X_pca",
    threshold: float = 0.01,
    radius: int = 50,
    method: str = "mean",
    threshold_spots: int = 5,
    use_sme: bool = False,
    reverse: bool = False,
    pseudotime_key: str = "dpt_pseudotime",
    max_nodes: int = 4,
    run_knn: bool = False,
    copy: bool = False,
) -> Optional[AnnData]:

    """\
    Perform pseudotime analysis.

    Parameters
    ----------
    adata
        Annotated data matrix.
    use_label
        Use label result of cluster method.
    eps
        The maximum distance between two samples for one to be considered as
        in the neighborhood of the other. This is not a maximum bound on the
        distances of points within a cluster. This is the most important DBSCAN
        parameter to choose appropriately for your data set and distance function.
    threshold
        Threshold to find the significant connection for PAGA graph.
    radius
        radius to adjust data for diffusion map
    method
        method to adjust the data.
    use_sme
        Use adjusted feature by SME normalization or not
    reverse
        Reverse the pseudotime score
    pseudotime_key
        Key to store pseudotime
    max_nodes
        Maximum number of node in available paths
    copy
        Return a copy instead of writing to adata.
    Returns
    -------
    Anndata
    """

    try:
        del adata.obsm["X_diffmap"]
    except:
        pass
    try:
        del adata.obsm["X_draw_graph_fr"]
    except:
        pass

    assert use_label != None, "Please choose the right `use_label`!"

    # Localize
    from stlearn.spatials.clustering import localization

    if "sub_clusters_laber" not in adata.obs.columns:
        localization(adata, use_label=use_label, eps=eps)

    # Running knn

    if run_knn:
        from stlearn.pp import neighbors

        neighbors(adata, n_neighbors=n_neighbors, use_rep=use_rep, random_state=0)

    # Running paga
    scanpy.tl.paga(adata, groups=use_label)

    # Denoising the graph
    scanpy.tl.diffmap(adata)

    if use_sme:
        from stlearn.spatials.morphology import adjust

        adjust(adata, use_data="X_diffmap", radius=radius, method=method)
        adata.obsm["X_diffmap"] = adata.obsm["X_diffmap_morphology"]

    # Get connection matrix
    cnt_matrix = adata.uns["paga"]["connectivities"].toarray()

    # Filter by threshold

    cnt_matrix[cnt_matrix < threshold] = 0.0
    cnt_matrix = pd.DataFrame(cnt_matrix)

    # Mapping louvain label to subcluster

    cat_ind = adata.uns[use_label + "_index_dict"]

    split_node = {}
    for label in adata.obs[use_label].unique():
        meaningful_sub = []
        for i in adata.obs[adata.obs[use_label] == label][
            "sub_cluster_labels"
        ].unique():
            if (
                len(adata.obs[adata.obs["sub_cluster_labels"] == str(i)])
                > threshold_spots
            ):
                meaningful_sub.append(i)

        split_node[cat_ind[label]] = meaningful_sub

    adata.uns["threshold_spots"] = threshold_spots
    adata.uns["split_node"] = split_node

    # Replicate louvain label row to prepare for subcluster connection
    # matrix construction
    replicate_list = np.array([])
    for i in range(0, len(cnt_matrix)):
        replicate_list = np.concatenate(
            [replicate_list, np.array([i] * len(split_node[i]))]
        )

    # Connection matrix for subcluster
    cnt_matrix = cnt_matrix.loc[replicate_list.astype(int), replicate_list.astype(int)]

    # Replace column and index
    cnt_matrix.columns = replace_with_dict(cnt_matrix.columns, split_node)
    cnt_matrix.index = replace_with_dict(cnt_matrix.index, split_node)

    # Sort column and index
    cnt_matrix = cnt_matrix.loc[
        selection_sort(np.array(cnt_matrix.columns)),
        selection_sort(np.array(cnt_matrix.index)),
    ]

    # Create a connection graph of subclusters
    G = nx.from_pandas_adjacency(cnt_matrix)
    G_nodes = list(range(len(G.nodes)))

    node_convert = {}
    for pair in zip(list(G.nodes), G_nodes):
        node_convert[pair[1]] = pair[0]

    adata.uns["global_graph"] = {}
    adata.uns["global_graph"]["graph"] = nx.to_scipy_sparse_array(G)
    adata.uns["global_graph"]["node_dict"] = node_convert

    # Create centroid dict for subclusters
    from sklearn.neighbors import NearestCentroid

    clf = NearestCentroid()
    clf.fit(adata.obs[["imagecol", "imagerow"]].values, adata.obs["sub_cluster_labels"])
    centroid_dict = dict(zip(clf.classes_.astype(int), clf.centroids_))

    def closest_node(node, nodes):
        nodes = np.asarray(nodes)
        dist_2 = np.sum((nodes - node) ** 2, axis=1)
        return np.argmin(dist_2)

    for cl in adata.obs["sub_cluster_labels"].unique():
        cl_points = adata.obs[adata.obs["sub_cluster_labels"] == cl][
            ["imagecol", "imagerow"]
        ].values
        new_centroid = cl_points[closest_node(centroid_dict[int(cl)], cl_points)]
        centroid_dict[int(cl)] = new_centroid

    adata.uns["centroid_dict"] = centroid_dict
    centroid_dict = {int(key): centroid_dict[key] for key in centroid_dict}

    # Running diffusion pseudo-time
    scanpy.tl.dpt(adata)

    if reverse:
        adata.obs[pseudotime_key] = 1 - adata.obs[pseudotime_key]

    store_available_paths(adata, threshold, use_label, max_nodes, pseudotime_key)

    return adata if copy else None


######## utils ########


def replace_with_dict(ar, dic):
    # Extract out keys and values
    k = np.array(list(dic.keys()), dtype=object)
    v = np.array(list(dic.values()), dtype=object)

    out = np.zeros_like(ar)
    for key, val in zip(k, v):
        out[ar == key] = val
    return out


def selection_sort(x):
    for i in range(len(x)):
        swap = i + np.argmin(x[i:])
        (x[i], x[swap]) = (x[swap], x[i])
    return x


def store_available_paths(adata, threshold, use_label, max_nodes, pseudotime_key):

    # Read original PAGA graph
    G = nx.from_numpy_array(adata.uns["paga"]["connectivities"].toarray())
    edge_weights = nx.get_edge_attributes(G, "weight")
    G.remove_edges_from((e for e, w in edge_weights.items() if w < threshold))

    H = G.to_directed()

    # Calculate pseudotime for each node
    node_pseudotime = {}

    for node in H.nodes:
        node_pseudotime[node] = adata.obs.query(use_label + " == '" + str(node) + "'")[
            pseudotime_key
        ].max()

    # Force original PAGA to directed PAGA based on pseudotime
    edge_to_remove = []
    for edge in H.edges:
        if node_pseudotime[edge[0]] - node_pseudotime[edge[1]] > 0:
            edge_to_remove.append(edge)
    H.remove_edges_from(edge_to_remove)

    # Extract all available paths
    all_paths = {}

    for source in H.nodes:
        for target in H.nodes:
            paths = nx.all_simple_paths(H, source=source, target=target)
            for i, path in enumerate(paths):
                if len(path) < max_nodes:
                    all_paths[i] = path

    # all_paths = list(map(lambda x: " - ".join(np.array(x).astype(str)),all_paths))

    adata.uns["available_paths"] = all_paths
    print(
        "All available trajectory paths are stored in adata.uns['available_paths'] with length < "
        + str(max_nodes)
        + " nodes"
    )



from pathlib import Path
import matplotlib
from matplotlib import pyplot as plt
import numpy as np
from typing import Optional, Union
from anndata import AnnData
import os
from stlearn._compat import Literal


def add_mask(
    adata: AnnData,
    imgpath: Union[Path, str],
    key: str = "mask",
    copy: bool = False,
) -> Optional[AnnData]:
    """\
    Adding binary mask image to the Anndata object

    Parameters
    ----------
    adata
        Anndata object.
    imgpath
        Image mask path.
    key
        Label for mask.
    copy
        Return a copy instead of writing to adata.
    Returns
    -------
    Depending on `copy`, returns or updates `adata` with the following fields.
    **mask_image** : `adata.uns` field
        Array format of image, saving by Pillow package.
    """
    try:
        library_id = list(adata.uns["spatial"].keys())[0]
        quality = adata.uns["spatial"][library_id]["use_quality"]
    except:
        raise KeyError(
            f"""\
        Please read ST data first and try again
        """
        )

    if imgpath is not None and os.path.isfile(imgpath):
        try:
            img = plt.imread(imgpath, 0)
            assert (
                img.shape == adata.uns["spatial"][library_id]["images"][quality].shape
            ), "\
            size of mask image does not match size of H&E images"
            if "mask_image" not in adata.uns:
                adata.uns["mask_image"] = {}
            if library_id not in adata.uns["mask_image"]:
                adata.uns["mask_image"][library_id] = {}
            if key not in adata.uns:
                adata.uns["mask_image"][library_id][key] = {}

            adata.uns["mask_image"][library_id][key][quality] = img
            print("Added tissue mask to the object!")

            return adata if copy else None
        except:
            raise ValueError(
                f"""\
            {imgpath!r} does not end on a valid extension.
            """
            )
    else:
        raise ValueError(
            f"""\
        {imgpath!r} does not end on a valid extension.
        """
        )
    return adata if copy else None


def apply_mask(
    adata: AnnData,
    masks: Optional[list] = "all",
    select: str = "black",
    cmap: str = "default",
    copy: bool = False,
) -> Optional[AnnData]:
    """\
    Parsing the old spaital transcriptomics data

    Parameters
    ----------
    adata
        Anndata object.
    masks
        List of masks for parsing
        if `all`, apply on all masks
    select
        color for ROI, choose from `black` and `white`.
    cmap
        Color map to use.
    copy
        Return a copy instead of writing to adata.
    Returns
    -------
    Depending on `copy`, returns or updates `adata` with the following fields.
    **mask_image** : `adata.uns` field
        Array format of image, saving by Pillow package.
    """
    from scanpy.plotting import palettes
    from stlearn.plotting import palettes_st

    if cmap == "vega_10_scanpy":
        cmap = palettes.vega_10_scanpy
    elif cmap == "vega_20_scanpy":
        cmap = palettes.vega_20_scanpy
    elif cmap == "default_102":
        cmap = palettes.default_102
    elif cmap == "default_28":
        cmap = palettes.default_28
    elif cmap == "jana_40":
        cmap = palettes_st.jana_40
    elif cmap == "default":
        cmap = palettes_st.default
    else:
        raise ValueError(
            "We only support vega_10_scanpy, vega_20_scanpy, default_28, default_102"
        )

    cmaps = matplotlib.colors.LinearSegmentedColormap.from_list("", cmap)

    cmap_ = plt.cm.get_cmap(cmaps)

    try:
        library_id = list(adata.uns["spatial"].keys())[0]
        quality = adata.uns["spatial"][library_id]["use_quality"]
    except:
        raise KeyError(
            f"""\
        Please read ST data first and try again
        """
        )

    if masks == "all":
        masks = list(adata.uns["mask_image"][library_id].keys())
    else:
        assert isinstance(masks, list)
    key = "mask_annotation"
    adata.obs[key] = "unassigned"
    adata.obs[key + "_code"] = -1
    mask_final = np.zeros(adata.uns["spatial"][library_id]["images"][quality].shape)
    for i, mask in enumerate(masks):
        try:
            mask_image = adata.uns["mask_image"][library_id][mask][quality]
        except:
            raise KeyError(
                f"""\
            Please load mask {mask} images first and try again
            """
            )

        if select == "black":
            mask_image = np.where(mask_image > 155, 0, 1)
        elif select == "white":
            mask_image = np.where(mask_image > 155, 0, 1)
        else:
            raise ValueError(
                f"""\
            Only support black and white mask yet.
            """
            )
        mask_image_2d = mask_image.mean(axis=2)
        apply_spot_mask = (
            lambda x: [i, mask]
            if mask_image_2d[int(x["imagerow"]), int(x["imagecol"])] == 1
            else [x[key + "_code"], x[key]]
        )
        spot_mask_df = adata.obs.apply(apply_spot_mask, axis=1, result_type="expand")
        adata.obs[key + "_code"] = spot_mask_df[0]
        adata.obs[key] = spot_mask_df[1]

        c = cmap_(int(i) / (len(cmap) - 1))
        mask_final = mask_final + mask_image * np.array(c[0:3])
    print(f"""Mask annotation for spots added to `adata.obs["{key}"]`""")
    mask_final[mask_final == 0] = 1
    adata.uns[key] = mask_final
    print(f"""Mask annotation for H&E image added to `adata.uns["{key}"]`""")
    return adata if copy else None
